# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from . import simulator_pb2 as simulator__pb2

GRPC_GENERATED_VERSION = '1.76.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + ' but the generated code in simulator_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class SimulatorServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.HealthCheck = channel.unary_unary(
                '/pulsim.api.v1.SimulatorService/HealthCheck',
                request_serializer=simulator__pb2.HealthCheckRequest.SerializeToString,
                response_deserializer=simulator__pb2.HealthCheckResponse.FromString,
                _registered_method=True)
        self.CreateSession = channel.unary_unary(
                '/pulsim.api.v1.SimulatorService/CreateSession',
                request_serializer=simulator__pb2.CreateSessionRequest.SerializeToString,
                response_deserializer=simulator__pb2.CreateSessionResponse.FromString,
                _registered_method=True)
        self.ListSessions = channel.unary_unary(
                '/pulsim.api.v1.SimulatorService/ListSessions',
                request_serializer=simulator__pb2.ListSessionsRequest.SerializeToString,
                response_deserializer=simulator__pb2.ListSessionsResponse.FromString,
                _registered_method=True)
        self.GetSession = channel.unary_unary(
                '/pulsim.api.v1.SimulatorService/GetSession',
                request_serializer=simulator__pb2.GetSessionRequest.SerializeToString,
                response_deserializer=simulator__pb2.GetSessionResponse.FromString,
                _registered_method=True)
        self.StartSimulation = channel.unary_unary(
                '/pulsim.api.v1.SimulatorService/StartSimulation',
                request_serializer=simulator__pb2.StartSimulationRequest.SerializeToString,
                response_deserializer=simulator__pb2.StartSimulationResponse.FromString,
                _registered_method=True)
        self.PauseSimulation = channel.unary_unary(
                '/pulsim.api.v1.SimulatorService/PauseSimulation',
                request_serializer=simulator__pb2.PauseSimulationRequest.SerializeToString,
                response_deserializer=simulator__pb2.PauseSimulationResponse.FromString,
                _registered_method=True)
        self.ResumeSimulation = channel.unary_unary(
                '/pulsim.api.v1.SimulatorService/ResumeSimulation',
                request_serializer=simulator__pb2.ResumeSimulationRequest.SerializeToString,
                response_deserializer=simulator__pb2.ResumeSimulationResponse.FromString,
                _registered_method=True)
        self.StopSimulation = channel.unary_unary(
                '/pulsim.api.v1.SimulatorService/StopSimulation',
                request_serializer=simulator__pb2.StopSimulationRequest.SerializeToString,
                response_deserializer=simulator__pb2.StopSimulationResponse.FromString,
                _registered_method=True)
        self.StreamWaveforms = channel.unary_stream(
                '/pulsim.api.v1.SimulatorService/StreamWaveforms',
                request_serializer=simulator__pb2.StreamWaveformsRequest.SerializeToString,
                response_deserializer=simulator__pb2.WaveformStreamResponse.FromString,
                _registered_method=True)
        self.GetResult = channel.unary_unary(
                '/pulsim.api.v1.SimulatorService/GetResult',
                request_serializer=simulator__pb2.GetResultRequest.SerializeToString,
                response_deserializer=simulator__pb2.GetResultResponse.FromString,
                _registered_method=True)
        self.DownloadResult = channel.unary_stream(
                '/pulsim.api.v1.SimulatorService/DownloadResult',
                request_serializer=simulator__pb2.DownloadResultRequest.SerializeToString,
                response_deserializer=simulator__pb2.DownloadResultResponse.FromString,
                _registered_method=True)
        self.UploadModel = channel.unary_unary(
                '/pulsim.api.v1.SimulatorService/UploadModel',
                request_serializer=simulator__pb2.UploadModelRequest.SerializeToString,
                response_deserializer=simulator__pb2.UploadModelResponse.FromString,
                _registered_method=True)
        self.ListModels = channel.unary_unary(
                '/pulsim.api.v1.SimulatorService/ListModels',
                request_serializer=simulator__pb2.ListModelsRequest.SerializeToString,
                response_deserializer=simulator__pb2.ListModelsResponse.FromString,
                _registered_method=True)
        self.GetModel = channel.unary_unary(
                '/pulsim.api.v1.SimulatorService/GetModel',
                request_serializer=simulator__pb2.GetModelRequest.SerializeToString,
                response_deserializer=simulator__pb2.GetModelResponse.FromString,
                _registered_method=True)
        self.DeleteModel = channel.unary_unary(
                '/pulsim.api.v1.SimulatorService/DeleteModel',
                request_serializer=simulator__pb2.DeleteModelRequest.SerializeToString,
                response_deserializer=simulator__pb2.DeleteModelResponse.FromString,
                _registered_method=True)
        self.CreateSweep = channel.unary_unary(
                '/pulsim.api.v1.SimulatorService/CreateSweep',
                request_serializer=simulator__pb2.CreateSweepRequest.SerializeToString,
                response_deserializer=simulator__pb2.CreateSweepResponse.FromString,
                _registered_method=True)
        self.RunSweep = channel.unary_unary(
                '/pulsim.api.v1.SimulatorService/RunSweep',
                request_serializer=simulator__pb2.RunSweepRequest.SerializeToString,
                response_deserializer=simulator__pb2.RunSweepResponse.FromString,
                _registered_method=True)
        self.GetSweepResults = channel.unary_unary(
                '/pulsim.api.v1.SimulatorService/GetSweepResults',
                request_serializer=simulator__pb2.GetSweepResultsRequest.SerializeToString,
                response_deserializer=simulator__pb2.GetSweepResultsResponse.FromString,
                _registered_method=True)


class SimulatorServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def HealthCheck(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateSession(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListSessions(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetSession(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StartSimulation(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PauseSimulation(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ResumeSimulation(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StopSimulation(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StreamWaveforms(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetResult(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DownloadResult(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UploadModel(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListModels(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetModel(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteModel(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateSweep(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RunSweep(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetSweepResults(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_SimulatorServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'HealthCheck': grpc.unary_unary_rpc_method_handler(
                    servicer.HealthCheck,
                    request_deserializer=simulator__pb2.HealthCheckRequest.FromString,
                    response_serializer=simulator__pb2.HealthCheckResponse.SerializeToString,
            ),
            'CreateSession': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateSession,
                    request_deserializer=simulator__pb2.CreateSessionRequest.FromString,
                    response_serializer=simulator__pb2.CreateSessionResponse.SerializeToString,
            ),
            'ListSessions': grpc.unary_unary_rpc_method_handler(
                    servicer.ListSessions,
                    request_deserializer=simulator__pb2.ListSessionsRequest.FromString,
                    response_serializer=simulator__pb2.ListSessionsResponse.SerializeToString,
            ),
            'GetSession': grpc.unary_unary_rpc_method_handler(
                    servicer.GetSession,
                    request_deserializer=simulator__pb2.GetSessionRequest.FromString,
                    response_serializer=simulator__pb2.GetSessionResponse.SerializeToString,
            ),
            'StartSimulation': grpc.unary_unary_rpc_method_handler(
                    servicer.StartSimulation,
                    request_deserializer=simulator__pb2.StartSimulationRequest.FromString,
                    response_serializer=simulator__pb2.StartSimulationResponse.SerializeToString,
            ),
            'PauseSimulation': grpc.unary_unary_rpc_method_handler(
                    servicer.PauseSimulation,
                    request_deserializer=simulator__pb2.PauseSimulationRequest.FromString,
                    response_serializer=simulator__pb2.PauseSimulationResponse.SerializeToString,
            ),
            'ResumeSimulation': grpc.unary_unary_rpc_method_handler(
                    servicer.ResumeSimulation,
                    request_deserializer=simulator__pb2.ResumeSimulationRequest.FromString,
                    response_serializer=simulator__pb2.ResumeSimulationResponse.SerializeToString,
            ),
            'StopSimulation': grpc.unary_unary_rpc_method_handler(
                    servicer.StopSimulation,
                    request_deserializer=simulator__pb2.StopSimulationRequest.FromString,
                    response_serializer=simulator__pb2.StopSimulationResponse.SerializeToString,
            ),
            'StreamWaveforms': grpc.unary_stream_rpc_method_handler(
                    servicer.StreamWaveforms,
                    request_deserializer=simulator__pb2.StreamWaveformsRequest.FromString,
                    response_serializer=simulator__pb2.WaveformStreamResponse.SerializeToString,
            ),
            'GetResult': grpc.unary_unary_rpc_method_handler(
                    servicer.GetResult,
                    request_deserializer=simulator__pb2.GetResultRequest.FromString,
                    response_serializer=simulator__pb2.GetResultResponse.SerializeToString,
            ),
            'DownloadResult': grpc.unary_stream_rpc_method_handler(
                    servicer.DownloadResult,
                    request_deserializer=simulator__pb2.DownloadResultRequest.FromString,
                    response_serializer=simulator__pb2.DownloadResultResponse.SerializeToString,
            ),
            'UploadModel': grpc.unary_unary_rpc_method_handler(
                    servicer.UploadModel,
                    request_deserializer=simulator__pb2.UploadModelRequest.FromString,
                    response_serializer=simulator__pb2.UploadModelResponse.SerializeToString,
            ),
            'ListModels': grpc.unary_unary_rpc_method_handler(
                    servicer.ListModels,
                    request_deserializer=simulator__pb2.ListModelsRequest.FromString,
                    response_serializer=simulator__pb2.ListModelsResponse.SerializeToString,
            ),
            'GetModel': grpc.unary_unary_rpc_method_handler(
                    servicer.GetModel,
                    request_deserializer=simulator__pb2.GetModelRequest.FromString,
                    response_serializer=simulator__pb2.GetModelResponse.SerializeToString,
            ),
            'DeleteModel': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteModel,
                    request_deserializer=simulator__pb2.DeleteModelRequest.FromString,
                    response_serializer=simulator__pb2.DeleteModelResponse.SerializeToString,
            ),
            'CreateSweep': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateSweep,
                    request_deserializer=simulator__pb2.CreateSweepRequest.FromString,
                    response_serializer=simulator__pb2.CreateSweepResponse.SerializeToString,
            ),
            'RunSweep': grpc.unary_unary_rpc_method_handler(
                    servicer.RunSweep,
                    request_deserializer=simulator__pb2.RunSweepRequest.FromString,
                    response_serializer=simulator__pb2.RunSweepResponse.SerializeToString,
            ),
            'GetSweepResults': grpc.unary_unary_rpc_method_handler(
                    servicer.GetSweepResults,
                    request_deserializer=simulator__pb2.GetSweepResultsRequest.FromString,
                    response_serializer=simulator__pb2.GetSweepResultsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'pulsim.api.v1.SimulatorService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('pulsim.api.v1.SimulatorService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class SimulatorService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def HealthCheck(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/pulsim.api.v1.SimulatorService/HealthCheck',
            simulator__pb2.HealthCheckRequest.SerializeToString,
            simulator__pb2.HealthCheckResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CreateSession(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/pulsim.api.v1.SimulatorService/CreateSession',
            simulator__pb2.CreateSessionRequest.SerializeToString,
            simulator__pb2.CreateSessionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListSessions(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/pulsim.api.v1.SimulatorService/ListSessions',
            simulator__pb2.ListSessionsRequest.SerializeToString,
            simulator__pb2.ListSessionsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetSession(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/pulsim.api.v1.SimulatorService/GetSession',
            simulator__pb2.GetSessionRequest.SerializeToString,
            simulator__pb2.GetSessionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def StartSimulation(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/pulsim.api.v1.SimulatorService/StartSimulation',
            simulator__pb2.StartSimulationRequest.SerializeToString,
            simulator__pb2.StartSimulationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def PauseSimulation(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/pulsim.api.v1.SimulatorService/PauseSimulation',
            simulator__pb2.PauseSimulationRequest.SerializeToString,
            simulator__pb2.PauseSimulationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ResumeSimulation(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/pulsim.api.v1.SimulatorService/ResumeSimulation',
            simulator__pb2.ResumeSimulationRequest.SerializeToString,
            simulator__pb2.ResumeSimulationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def StopSimulation(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/pulsim.api.v1.SimulatorService/StopSimulation',
            simulator__pb2.StopSimulationRequest.SerializeToString,
            simulator__pb2.StopSimulationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def StreamWaveforms(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/pulsim.api.v1.SimulatorService/StreamWaveforms',
            simulator__pb2.StreamWaveformsRequest.SerializeToString,
            simulator__pb2.WaveformStreamResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetResult(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/pulsim.api.v1.SimulatorService/GetResult',
            simulator__pb2.GetResultRequest.SerializeToString,
            simulator__pb2.GetResultResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DownloadResult(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/pulsim.api.v1.SimulatorService/DownloadResult',
            simulator__pb2.DownloadResultRequest.SerializeToString,
            simulator__pb2.DownloadResultResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def UploadModel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/pulsim.api.v1.SimulatorService/UploadModel',
            simulator__pb2.UploadModelRequest.SerializeToString,
            simulator__pb2.UploadModelResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListModels(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/pulsim.api.v1.SimulatorService/ListModels',
            simulator__pb2.ListModelsRequest.SerializeToString,
            simulator__pb2.ListModelsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetModel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/pulsim.api.v1.SimulatorService/GetModel',
            simulator__pb2.GetModelRequest.SerializeToString,
            simulator__pb2.GetModelResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DeleteModel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/pulsim.api.v1.SimulatorService/DeleteModel',
            simulator__pb2.DeleteModelRequest.SerializeToString,
            simulator__pb2.DeleteModelResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CreateSweep(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/pulsim.api.v1.SimulatorService/CreateSweep',
            simulator__pb2.CreateSweepRequest.SerializeToString,
            simulator__pb2.CreateSweepResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RunSweep(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/pulsim.api.v1.SimulatorService/RunSweep',
            simulator__pb2.RunSweepRequest.SerializeToString,
            simulator__pb2.RunSweepResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetSweepResults(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/pulsim.api.v1.SimulatorService/GetSweepResults',
            simulator__pb2.GetSweepResultsRequest.SerializeToString,
            simulator__pb2.GetSweepResultsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



"""Analytical electrothermal regression tests for thermal-enabled component types."""

from __future__ import annotations

import math
import statistics
import textwrap
from dataclasses import dataclass

import pulsim as ps
import pytest


AMBIENT_C = 25.0
RTH = 2.0
CTH = 1e-3


@dataclass(frozen=True)
class ThermalAnalyticCase:
    component_type: str
    component_name: str
    expected_power_w: float
    components_yaml: str


THERMAL_ANALYTIC_CASES: tuple[ThermalAnalyticCase, ...] = (
    ThermalAnalyticCase(
        component_type="resistor",
        component_name="R1",
        expected_power_w=2.5,  # V^2 / R = 5^2 / 10
        components_yaml="""
- type: voltage_source
  name: V1
  nodes: [n, 0]
  waveform: {type: dc, value: 5}
- type: resistor
  name: R1
  nodes: [n, 0]
  value: 10
  thermal:
    enabled: true
    rth: 2
    cth: 1e-3
""",
    ),
    ThermalAnalyticCase(
        component_type="diode",
        component_name="D1",
        expected_power_w=1.25,  # g_on * V^2 = 0.05 * 5^2
        components_yaml="""
- type: voltage_source
  name: V1
  nodes: [n, 0]
  waveform: {type: dc, value: 5}
- type: diode
  name: D1
  nodes: [n, 0]
  g_on: 0.05
  g_off: 1e-9
  thermal:
    enabled: true
    rth: 2
    cth: 1e-3
""",
    ),
    ThermalAnalyticCase(
        component_type="mosfet",
        component_name="M1",
        expected_power_w=1.25,  # forced-on branch: kp * Vds^2 = 0.05 * 5^2
        components_yaml="""
- type: voltage_source
  name: VDS
  nodes: [d, 0]
  waveform: {type: dc, value: 5}
- type: voltage_source
  name: VG
  nodes: [g, 0]
  waveform: {type: dc, value: 0}
- type: mosfet
  name: M1
  nodes: [g, d, 0]
  vth: 2
  kp: 0.05
  lambda: 0
  thermal:
    enabled: true
    rth: 2
    cth: 1e-3
- type: pwm_generator
  name: PWM_M1
  nodes: [g]
  frequency: 1000
  duty: 1.0
  target_component: M1
""",
    ),
    ThermalAnalyticCase(
        component_type="igbt",
        component_name="Q1",
        expected_power_w=1.25,  # forced-on branch: g_on * Vce^2 = 0.05 * 5^2
        components_yaml="""
- type: voltage_source
  name: VCE
  nodes: [c, 0]
  waveform: {type: dc, value: 5}
- type: voltage_source
  name: VG
  nodes: [g, 0]
  waveform: {type: dc, value: 0}
- type: igbt
  name: Q1
  nodes: [g, c, 0]
  vth: 2
  g_on: 0.05
  g_off: 1e-9
  thermal:
    enabled: true
    rth: 2
    cth: 1e-3
- type: pwm_generator
  name: PWM_Q1
  nodes: [g]
  frequency: 1000
  duty: 1.0
  target_component: Q1
""",
    ),
    ThermalAnalyticCase(
        component_type="bjt_npn",
        component_name="QN1",
        expected_power_w=0.5,  # surrogate kp=beta*1e-3 => 0.02*5^2 when forced-on
        components_yaml="""
- type: voltage_source
  name: VC
  nodes: [c, 0]
  waveform: {type: dc, value: 5}
- type: voltage_source
  name: VB
  nodes: [b, 0]
  waveform: {type: dc, value: 0}
- type: bjt_npn
  name: QN1
  nodes: [b, c, 0]
  beta: 20
  thermal:
    enabled: true
    rth: 2
    cth: 1e-3
- type: pwm_generator
  name: PWM_QN1
  nodes: [b]
  frequency: 1000
  duty: 1.0
  target_component: QN1
""",
    ),
    ThermalAnalyticCase(
        component_type="bjt_pnp",
        component_name="QP1",
        expected_power_w=0.5,  # surrogate kp=beta*1e-3 => 0.02*|V|^2 when forced-on
        components_yaml="""
- type: voltage_source
  name: VE
  nodes: [e, 0]
  waveform: {type: dc, value: 5}
- type: voltage_source
  name: VB
  nodes: [b, 0]
  waveform: {type: dc, value: 0}
- type: bjt_pnp
  name: QP1
  nodes: [b, 0, e]
  beta: 20
  thermal:
    enabled: true
    rth: 2
    cth: 1e-3
- type: pwm_generator
  name: PWM_QP1
  nodes: [b]
  frequency: 1000
  duty: 1.0
  target_component: QP1
""",
    ),
)

EXPECTED_THERMAL_ENABLED_TYPES = {
    "resistor",
    "diode",
    "mosfet",
    "igbt",
    "bjt_npn",
    "bjt_pnp",
}


def _build_yaml(case: ThermalAnalyticCase) -> str:
    simulation = textwrap.dedent(
        """
        schema: pulsim-v1
        version: 1
        simulation:
          tstart: 0
          tstop: 5e-4
          dt: 1e-6
          enable_losses: true
          thermal:
            enabled: true
            ambient: 25
            policy: loss_only
            default_rth: 2
            default_cth: 1e-3
        components:
        """
    ).strip()
    body = textwrap.indent(textwrap.dedent(case.components_yaml).strip(), "  ")
    return f"{simulation}\n{body}\n"


def _expected_thermal_trace_from_constant_power(
    power_w: float,
    time: list[float],
    ambient_c: float,
    rth: float,
    cth: float,
) -> list[float]:
    trace = [ambient_c]
    current = ambient_c
    for i in range(1, len(time)):
        dt = float(time[i]) - float(time[i - 1])
        if cth <= 0.0:
            current = ambient_c + power_w * rth
        else:
            tau = max(rth * cth, 1e-18)
            delta = current - ambient_c
            a = math.exp(-dt / tau)
            current = ambient_c + delta * a + (power_w * rth) * (1.0 - a)
        trace.append(current)
    return trace


def _assert_close(value: float, expected: float, *, rel: float = 1e-9, abs_: float = 1e-9) -> None:
    assert math.isclose(value, expected, rel_tol=rel, abs_tol=abs_), (
        f"value={value} expected={expected} rel={rel} abs={abs_}"
    )


def test_thermal_analytic_cases_cover_all_supported_thermal_types() -> None:
    covered = {case.component_type for case in THERMAL_ANALYTIC_CASES}
    assert covered == EXPECTED_THERMAL_ENABLED_TYPES


@pytest.mark.parametrize("case", THERMAL_ANALYTIC_CASES, ids=lambda case: case.component_type)
def test_thermal_response_matches_discrete_rc_theory(case: ThermalAnalyticCase) -> None:
    parser = ps.YamlParser()
    circuit, options = parser.load_string(_build_yaml(case))
    assert parser.errors == [], parser.errors

    options.adaptive_timestep = False
    options.dt_min = options.dt
    options.dt_max = options.dt

    result = ps.Simulator(circuit, options).run_transient(circuit.initial_state())
    assert result.success

    channel = f"T({case.component_name})"
    assert channel in result.virtual_channels
    trace = [float(v) for v in result.virtual_channels[channel]]
    time = [float(t) for t in result.time]
    assert len(trace) == len(time)
    assert len(trace) >= 2
    assert all(curr >= prev - 1e-12 for prev, curr in zip(trace, trace[1:]))

    expected_trace = _expected_thermal_trace_from_constant_power(
        case.expected_power_w, time, AMBIENT_C, RTH, CTH
    )
    assert len(expected_trace) == len(trace)

    expected_final = expected_trace[-1]
    expected_peak = max(expected_trace)
    expected_avg = statistics.fmean(expected_trace)

    _assert_close(trace[-1], expected_final)
    _assert_close(max(trace), expected_peak)
    _assert_close(statistics.fmean(trace), expected_avg)

    component_rows = {item.component_name: item for item in result.component_electrothermal}
    assert case.component_name in component_rows
    component_row = component_rows[case.component_name]
    assert component_row.thermal_enabled
    _assert_close(float(component_row.average_power), case.expected_power_w)
    _assert_close(float(component_row.final_temperature), trace[-1])
    _assert_close(float(component_row.peak_temperature), max(trace))
    _assert_close(float(component_row.average_temperature), statistics.fmean(trace))

    summary_rows = {item.device_name: item for item in result.thermal_summary.device_temperatures}
    assert case.component_name in summary_rows
    summary_row = summary_rows[case.component_name]
    _assert_close(float(summary_row.final_temperature), trace[-1])
    _assert_close(float(summary_row.peak_temperature), max(trace))
    _assert_close(float(summary_row.average_temperature), statistics.fmean(trace))



"""Waveform post-processing pipeline for pulsim.

Provides deterministic, backend-owned metric computation for transient simulation
results. Supports time-domain metrics, spectral (FFT/harmonic/THD) analysis, and
power/efficiency measurements.

Public API:
    PostProcessingWindowMode    -- Window specification mode enum
    PostProcessingJobKind       -- Job type enum
    PostProcessingDiagnosticCode -- Diagnostic reason codes
    WindowFunction              -- Spectral window function enum
    PostProcessingWindowSpec    -- Window configuration dataclass
    PostProcessingJob           -- Single job configuration dataclass
    PostProcessingOptions       -- Collection of jobs to execute
    ScalarMetric                -- Named scalar metric result
    SpectralBin                 -- FFT frequency bin result
    HarmonicEntry               -- Harmonic component result
    UndefinedMetricEntry        -- Undefined metric with reason
    PostProcessingJobResult     -- Per-job result dataclass
    PostProcessingResult        -- Aggregated result dataclass
    run_post_processing         -- Execute jobs on a SimulationResult
    parse_post_processing_yaml  -- Parse simulation.post_processing YAML node
"""

from __future__ import annotations

import math
import time as _time_module
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional

import numpy as np


# ---------------------------------------------------------------------------
# Enumerations
# ---------------------------------------------------------------------------

class PostProcessingWindowMode(str, Enum):
    """Window specification mode for post-processing jobs."""
    Time = "time"    # bounds given as [t_start, t_end] in seconds
    Index = "index"  # bounds given as [i_start, i_end] sample indices
    Cycle = "cycle"  # bounds given as [cycle_start, cycle_end] with period


class PostProcessingJobKind(str, Enum):
    """Type of post-processing job."""
    TimeDomain = "time_domain"
    Spectral = "spectral"
    PowerEfficiency = "power_efficiency"


class PostProcessingDiagnosticCode(str, Enum):
    """Stable machine-readable reason codes for post-processing diagnostics."""
    Ok = "ok"
    InvalidConfiguration = "invalid_configuration"
    SignalNotFound = "signal_not_found"
    InvalidWindow = "invalid_window"
    InsufficientSamples = "insufficient_samples"
    SamplingMismatch = "sampling_mismatch"
    UndefinedMetric = "undefined_metric"
    NumericalFailure = "numerical_failure"


class WindowFunction(str, Enum):
    """Spectral window functions for FFT analysis."""
    Rectangular = "rectangular"
    Hann = "hann"
    Hamming = "hamming"
    Blackman = "blackman"
    FlatTop = "flat_top"


# ---------------------------------------------------------------------------
# Configuration dataclasses
# ---------------------------------------------------------------------------

@dataclass
class PostProcessingWindowSpec:
    """Window specification for a post-processing job.

    The active fields depend on *mode*:
      - Time:  t_start / t_end  (seconds)
      - Index: i_start / i_end  (sample indices, -1 means last)
      - Cycle: cycle_start / cycle_end / period  (cycle count + period seconds)
    """
    mode: PostProcessingWindowMode = PostProcessingWindowMode.Time
    t_start: float = 0.0
    t_end: float = float("inf")
    i_start: int = 0
    i_end: int = -1            # -1 → last sample
    cycle_start: int = 0
    cycle_end: int = -1        # -1 → cycle_start + 1
    period: float = 0.0        # required for Cycle mode
    min_samples: int = 4       # minimum samples required in window


@dataclass
class PostProcessingJob:
    """Configuration for a single post-processing job.

    Fields used depend on *kind*:
      - TimeDomain:      signals, window, metrics
      - Spectral:        signals[0], window, n_harmonics, fundamental_hz, window_function
      - PowerEfficiency: input/output voltage/current signals, window
    """
    job_id: str
    kind: PostProcessingJobKind
    signals: List[str] = field(default_factory=list)
    window: PostProcessingWindowSpec = field(default_factory=PostProcessingWindowSpec)

    # TimeDomain options
    metrics: List[str] = field(
        default_factory=lambda: ["rms", "mean", "min", "max", "p2p", "crest", "ripple"]
    )

    # Spectral options
    n_harmonics: int = 10
    fundamental_hz: float = 0.0   # 0 = auto-detect peak
    window_function: WindowFunction = WindowFunction.Rectangular

    # PowerEfficiency signal bindings
    input_voltage_signal: str = ""
    input_current_signal: str = ""
    output_voltage_signal: str = ""
    output_current_signal: str = ""


@dataclass
class PostProcessingOptions:
    """Collection of post-processing jobs to execute after simulation."""
    jobs: List[PostProcessingJob] = field(default_factory=list)


# ---------------------------------------------------------------------------
# Result dataclasses
# ---------------------------------------------------------------------------

@dataclass
class ScalarMetric:
    """A single named scalar metric value."""
    name: str
    value: float
    unit: str = ""
    domain: str = "time"
    source_signal: str = ""


@dataclass
class SpectralBin:
    """One FFT frequency bin."""
    frequency_hz: float
    magnitude: float
    phase_deg: float


@dataclass
class HarmonicEntry:
    """One harmonic component."""
    harmonic_number: int
    frequency_hz: float
    magnitude: float
    phase_deg: float
    magnitude_pct_fundamental: float


@dataclass
class UndefinedMetricEntry:
    """A metric that could not be computed, with stable reason code."""
    name: str
    reason: PostProcessingDiagnosticCode
    reason_message: str


@dataclass
class PostProcessingJobResult:
    """Result from a single post-processing job."""
    job_id: str
    kind: PostProcessingJobKind
    success: bool
    diagnostic: PostProcessingDiagnosticCode
    diagnostic_message: str = ""

    # TimeDomain results
    scalar_metrics: Dict[str, ScalarMetric] = field(default_factory=dict)

    # Spectral results
    spectrum_bins: List[SpectralBin] = field(default_factory=list)
    harmonics: List[HarmonicEntry] = field(default_factory=list)
    thd_pct: float = float("nan")
    fundamental_hz: float = float("nan")

    # PowerEfficiency results
    average_input_power: float = float("nan")
    average_output_power: float = float("nan")
    efficiency: float = float("nan")
    power_factor: float = float("nan")

    # Undefined metrics with reason codes
    undefined_metrics: List[UndefinedMetricEntry] = field(default_factory=list)

    # Metadata
    signal_names: List[str] = field(default_factory=list)
    window_i_start: int = 0
    window_i_end: int = 0
    sample_count: int = 0
    runtime_seconds: float = 0.0


@dataclass
class PostProcessingResult:
    """Aggregated result from all post-processing jobs."""
    success: bool
    jobs: List[PostProcessingJobResult] = field(default_factory=list)
    total_runtime_seconds: float = 0.0
    run_count: int = 1


# ---------------------------------------------------------------------------
# Internal helpers: window functions
# ---------------------------------------------------------------------------

def _make_window(kind: WindowFunction, n: int) -> np.ndarray:
    """Return a deterministic spectral window array of length *n*."""
    if n <= 0:
        return np.ones(0, dtype=np.float64)
    if kind == WindowFunction.Rectangular:
        return np.ones(n, dtype=np.float64)
    idx = np.arange(n, dtype=np.float64)
    m = max(n - 1, 1)
    if kind == WindowFunction.Hann:
        return 0.5 * (1.0 - np.cos(2.0 * np.pi * idx / m))
    if kind == WindowFunction.Hamming:
        return 0.54 - 0.46 * np.cos(2.0 * np.pi * idx / m)
    if kind == WindowFunction.Blackman:
        return (
            0.42
            - 0.5 * np.cos(2.0 * np.pi * idx / m)
            + 0.08 * np.cos(4.0 * np.pi * idx / m)
        )
    if kind == WindowFunction.FlatTop:
        a0, a1, a2, a3, a4 = (
            0.21557895, 0.41663158, 0.277263158, 0.083578947, 0.006947368
        )
        x = 2.0 * np.pi * idx / m
        return (
            a0
            - a1 * np.cos(x)
            + a2 * np.cos(2 * x)
            - a3 * np.cos(3 * x)
            + a4 * np.cos(4 * x)
        )
    return np.ones(n, dtype=np.float64)


def _window_coherent_gain(w: np.ndarray) -> float:
    """Return the coherent gain (mean) of a window array."""
    g = float(np.mean(w))
    return g if g != 0.0 else 1.0


# ---------------------------------------------------------------------------
# Internal helpers: window planning
# ---------------------------------------------------------------------------

def _resolve_window_indices(
    time_arr: np.ndarray,
    spec: PostProcessingWindowSpec,
) -> tuple[int, int, PostProcessingDiagnosticCode, str]:
    """Convert a PostProcessingWindowSpec to (i_start, i_end) sample bounds.

    Returns (i_start, i_end, code, message). i_end is exclusive.
    On error, returns (0, 0, error_code, description).
    """
    n = len(time_arr)
    if n == 0:
        return 0, 0, PostProcessingDiagnosticCode.InvalidWindow, "Empty time array"

    if spec.mode == PostProcessingWindowMode.Index:
        i0 = spec.i_start
        i1 = spec.i_end
        if i0 < 0:
            i0 = n + i0
        if i1 < 0:
            i1 = n + i1 + 1
        i0 = max(0, min(i0, n))
        i1 = max(i0, min(i1, n))
        if i1 <= i0:
            return 0, 0, PostProcessingDiagnosticCode.InvalidWindow, (
                f"Index window [{spec.i_start}, {spec.i_end}] is empty "
                f"after clamping to [{i0}, {i1}]"
            )
        return i0, i1, PostProcessingDiagnosticCode.Ok, ""

    if spec.mode == PostProcessingWindowMode.Time:
        t0 = spec.t_start
        t1 = spec.t_end if not math.isinf(spec.t_end) else float(time_arr[-1])
        if t1 < t0:
            return 0, 0, PostProcessingDiagnosticCode.InvalidWindow, (
                f"Time window t_start={t0} > t_end={t1}"
            )
        i0 = int(np.searchsorted(time_arr, t0, side="left"))
        i1 = int(np.searchsorted(time_arr, t1, side="right"))
        i0 = max(0, min(i0, n))
        i1 = max(i0, min(i1, n))
        if i1 <= i0:
            return 0, 0, PostProcessingDiagnosticCode.InvalidWindow, (
                f"Time window [{t0}, {t1}] contains no simulation samples"
            )
        return i0, i1, PostProcessingDiagnosticCode.Ok, ""

    if spec.mode == PostProcessingWindowMode.Cycle:
        if spec.period <= 0.0:
            return 0, 0, PostProcessingDiagnosticCode.InvalidWindow, (
                "Cycle window requires period > 0"
            )
        c_end = spec.cycle_end if spec.cycle_end >= 0 else spec.cycle_start
        t0 = spec.cycle_start * spec.period
        t1 = (c_end + 1) * spec.period
        t1 = min(t1, float(time_arr[-1]))
        if t1 <= t0:
            return 0, 0, PostProcessingDiagnosticCode.InvalidWindow, (
                f"Cycle window [{spec.cycle_start}, {c_end}] with "
                f"period={spec.period} yields empty time range"
            )
        i0 = int(np.searchsorted(time_arr, t0, side="left"))
        i1 = int(np.searchsorted(time_arr, t1, side="right"))
        i0 = max(0, min(i0, n))
        i1 = max(i0, min(i1, n))
        if i1 <= i0:
            return 0, 0, PostProcessingDiagnosticCode.InvalidWindow, (
                f"Cycle window [{spec.cycle_start}, {c_end}] contains no samples"
            )
        return i0, i1, PostProcessingDiagnosticCode.Ok, ""

    return 0, 0, PostProcessingDiagnosticCode.InvalidConfiguration, (
        f"Unknown window mode: {spec.mode}"
    )


# ---------------------------------------------------------------------------
# Internal helpers: metric engines
# ---------------------------------------------------------------------------

def _compute_time_domain_metrics(
    x: np.ndarray,
    requested: List[str],
    signal_name: str,
    unit: str = "",
) -> tuple[Dict[str, ScalarMetric], List[UndefinedMetricEntry]]:
    """Compute requested time-domain scalar metrics.

    Supported metric names (case-insensitive):
      rms, mean, min, max, p2p, crest, ripple, std
    """
    metrics: Dict[str, ScalarMetric] = {}
    undefined: List[UndefinedMetricEntry] = []

    if len(x) == 0:
        for name in requested:
            undefined.append(UndefinedMetricEntry(
                name=name,
                reason=PostProcessingDiagnosticCode.InsufficientSamples,
                reason_message="Empty signal window",
            ))
        return metrics, undefined

    mean_val = float(np.mean(x))
    rms_val = float(np.sqrt(np.mean(x ** 2)))
    min_val = float(np.min(x))
    max_val = float(np.max(x))
    p2p_val = max_val - min_val
    abs_max = max(abs(min_val), abs(max_val))

    for raw_name in requested:
        k = raw_name.lower().strip()
        if k == "rms":
            metrics["rms"] = ScalarMetric("rms", rms_val, unit, "time", signal_name)
        elif k == "mean":
            metrics["mean"] = ScalarMetric("mean", mean_val, unit, "time", signal_name)
        elif k in ("min", "minimum"):
            metrics["min"] = ScalarMetric("min", min_val, unit, "time", signal_name)
        elif k in ("max", "maximum"):
            metrics["max"] = ScalarMetric("max", max_val, unit, "time", signal_name)
        elif k in ("p2p", "peak_to_peak", "peak-to-peak", "pkpk"):
            metrics["p2p"] = ScalarMetric("p2p", p2p_val, unit, "time", signal_name)
        elif k in ("crest", "crest_factor"):
            if rms_val == 0.0:
                undefined.append(UndefinedMetricEntry(
                    "crest",
                    PostProcessingDiagnosticCode.UndefinedMetric,
                    "Crest factor undefined when RMS is zero",
                ))
            else:
                metrics["crest"] = ScalarMetric(
                    "crest", abs_max / rms_val, "", "time", signal_name
                )
        elif k == "ripple":
            # Ripple = peak-to-peak / |mean| (power-electronics convention).
            # Treat ripple as undefined when mean is negligible relative to RMS.
            # Threshold: |mean| < 0.1% of RMS (or RMS==0 edge case).
            rms_ref = max(rms_val, 1e-100)
            if abs(mean_val) / rms_ref < 1e-3:
                undefined.append(UndefinedMetricEntry(
                    "ripple",
                    PostProcessingDiagnosticCode.UndefinedMetric,
                    "Ripple undefined when mean is zero",
                ))
            else:
                metrics["ripple"] = ScalarMetric(
                    "ripple", p2p_val / abs(mean_val), "", "time", signal_name
                )
        elif k in ("std", "stddev", "standard_deviation"):
            metrics["std"] = ScalarMetric(
                "std", float(np.std(x)), unit, "time", signal_name
            )
        else:
            undefined.append(UndefinedMetricEntry(
                raw_name,
                PostProcessingDiagnosticCode.InvalidConfiguration,
                f"Unknown time-domain metric: '{raw_name}' "
                f"(accepted: rms, mean, min, max, p2p, crest, ripple, std)",
            ))

    return metrics, undefined


def _compute_spectral_metrics(
    x: np.ndarray,
    time_arr: np.ndarray,
    fundamental_hz: float,
    n_harmonics: int,
    window_kind: WindowFunction,
) -> tuple[
    List[SpectralBin],
    List[HarmonicEntry],
    float,
    float,
    List[UndefinedMetricEntry],
]:
    """Compute FFT spectrum, harmonics table, and THD.

    Returns (bins, harmonics, thd_pct, detected_fundamental_hz, undefined_metrics).
    """
    undefined: List[UndefinedMetricEntry] = []
    n = len(x)

    if n < 4:
        undefined.append(UndefinedMetricEntry(
            "spectrum",
            PostProcessingDiagnosticCode.InsufficientSamples,
            f"Spectral analysis requires >= 4 samples, got {n}",
        ))
        return [], [], float("nan"), float("nan"), undefined

    dt_arr = np.diff(time_arr)
    if len(dt_arr) == 0:
        undefined.append(UndefinedMetricEntry(
            "spectrum",
            PostProcessingDiagnosticCode.InsufficientSamples,
            "Need at least 2 time samples to determine sample rate",
        ))
        return [], [], float("nan"), float("nan"), undefined

    dt_mean = float(np.mean(dt_arr))
    if dt_mean <= 0.0:
        undefined.append(UndefinedMetricEntry(
            "spectrum",
            PostProcessingDiagnosticCode.SamplingMismatch,
            "Non-positive mean time step in window",
        ))
        return [], [], float("nan"), float("nan"), undefined

    # Apply spectral window
    w = _make_window(window_kind, n)
    cg = _window_coherent_gain(w)
    x_windowed = x * w

    # Compute FFT and one-sided amplitude spectrum
    X = np.fft.rfft(x_windowed)
    freqs = np.fft.rfftfreq(n, d=dt_mean)
    mag = np.abs(X) / (cg * n)

    # One-sided correction: double all bins except DC (index 0) and Nyquist
    mag_os = mag.copy()
    nyquist_idx = len(mag_os) - 1
    mag_os[1:nyquist_idx] *= 2.0

    phase_deg = np.degrees(np.angle(X))

    # Build spectrum bin list (exclude DC)
    bins: List[SpectralBin] = [
        SpectralBin(
            frequency_hz=float(freqs[i]),
            magnitude=float(mag_os[i]),
            phase_deg=float(phase_deg[i]),
        )
        for i in range(1, len(freqs))
    ]

    if not bins:
        undefined.append(UndefinedMetricEntry(
            "harmonics",
            PostProcessingDiagnosticCode.InsufficientSamples,
            "No spectral bins available",
        ))
        return bins, [], float("nan"), float("nan"), undefined

    freqs_arr = np.array([b.frequency_hz for b in bins], dtype=np.float64)
    mags_arr = np.array([b.magnitude for b in bins], dtype=np.float64)

    # Detect or locate fundamental
    if fundamental_hz <= 0.0:
        peak_idx = int(np.argmax(mags_arr))
        detected_f = float(freqs_arr[peak_idx])
    else:
        fundamental_bin_idx = int(np.argmin(np.abs(freqs_arr - fundamental_hz)))
        detected_f = float(freqs_arr[fundamental_bin_idx])

    f0_bin_idx = int(np.argmin(np.abs(freqs_arr - detected_f)))
    h1_mag = float(mags_arr[f0_bin_idx])

    if h1_mag == 0.0:
        undefined.append(UndefinedMetricEntry(
            "thd",
            PostProcessingDiagnosticCode.UndefinedMetric,
            "THD undefined: fundamental component magnitude is zero",
        ))
        return bins, [], float("nan"), detected_f, undefined

    # Build harmonic table
    harmonics: List[HarmonicEntry] = []
    harmonic_sum_sq = 0.0
    for h_num in range(1, n_harmonics + 1):
        target_f = detected_f * h_num
        h_idx = int(np.argmin(np.abs(freqs_arr - target_f)))
        h_mag = float(mags_arr[h_idx])
        h_phase = float(bins[h_idx].phase_deg)
        pct = h_mag / h1_mag * 100.0
        harmonics.append(HarmonicEntry(
            harmonic_number=h_num,
            frequency_hz=float(freqs_arr[h_idx]),
            magnitude=h_mag,
            phase_deg=h_phase,
            magnitude_pct_fundamental=pct,
        ))
        if h_num > 1:
            harmonic_sum_sq += h_mag ** 2

    thd_pct = math.sqrt(harmonic_sum_sq) / h1_mag * 100.0

    return bins, harmonics, thd_pct, detected_f, undefined


def _compute_power_efficiency_metrics(
    v_in: np.ndarray,
    i_in: np.ndarray,
    v_out: np.ndarray,
    i_out: np.ndarray,
) -> tuple[float, float, float, float, List[UndefinedMetricEntry]]:
    """Compute average power, efficiency, and power factor.

    Returns (p_in_W, p_out_W, efficiency_pct, power_factor, undefined_metrics).
    """
    undefined: List[UndefinedMetricEntry] = []

    def _avg_power(
        v: np.ndarray, i: np.ndarray, label: str
    ) -> tuple[float, Optional[str]]:
        if len(v) == 0 or len(i) == 0:
            return float("nan"), f"{label}: empty array"
        if len(v) != len(i):
            return float("nan"), (
                f"{label}: voltage/current length mismatch "
                f"({len(v)} vs {len(i)})"
            )
        return float(np.mean(v * i)), None

    p_in = float("nan")
    p_out = float("nan")
    efficiency = float("nan")
    pf = float("nan")

    if len(v_in) > 0 and len(i_in) > 0:
        p_in, err = _avg_power(v_in, i_in, "input")
        if err:
            undefined.append(UndefinedMetricEntry(
                "input_power", PostProcessingDiagnosticCode.InvalidConfiguration, err
            ))
            p_in = float("nan")

    if len(v_out) > 0 and len(i_out) > 0:
        p_out, err = _avg_power(v_out, i_out, "output")
        if err:
            undefined.append(UndefinedMetricEntry(
                "output_power", PostProcessingDiagnosticCode.InvalidConfiguration, err
            ))
            p_out = float("nan")

    if not math.isnan(p_in) and not math.isnan(p_out):
        if p_in > 0.0:
            efficiency = p_out / p_in * 100.0
        elif p_in == 0.0 and p_out == 0.0:
            efficiency = 100.0
        else:
            undefined.append(UndefinedMetricEntry(
                "efficiency",
                PostProcessingDiagnosticCode.UndefinedMetric,
                f"Efficiency undefined for P_in={p_in:.6g} W",
            ))

    # Power factor from input: P / S = P / (V_rms * I_rms)
    if len(v_in) > 0 and len(i_in) > 0 and not math.isnan(p_in):
        v_rms = float(np.sqrt(np.mean(v_in ** 2)))
        i_rms = float(np.sqrt(np.mean(i_in ** 2)))
        s = v_rms * i_rms
        if s > 0.0:
            pf = p_in / s
        else:
            undefined.append(UndefinedMetricEntry(
                "power_factor",
                PostProcessingDiagnosticCode.UndefinedMetric,
                "Power factor undefined: apparent power S = V_rms * I_rms is zero",
            ))

    return p_in, p_out, efficiency, pf, undefined


def _get_signal_data(
    channels: Dict[str, Any],
    signal_name: str,
) -> tuple[Optional[np.ndarray], PostProcessingDiagnosticCode, str]:
    """Resolve a named signal from the virtual_channels dict.

    Performs case-insensitive fallback lookup if exact name not found.
    Returns (array, code, message). On failure, array is None.
    """
    if not signal_name:
        return None, PostProcessingDiagnosticCode.InvalidConfiguration, (
            "Empty signal name"
        )
    if signal_name in channels:
        data = channels[signal_name]
        return np.asarray(data, dtype=np.float64), PostProcessingDiagnosticCode.Ok, ""
    # Case-insensitive fallback
    lower = signal_name.lower()
    for k, v in channels.items():
        if k.lower() == lower:
            return np.asarray(v, dtype=np.float64), PostProcessingDiagnosticCode.Ok, ""
    return None, PostProcessingDiagnosticCode.SignalNotFound, (
        f"Signal '{signal_name}' not found in result channels "
        f"(available: {list(channels.keys())})"
    )


# ---------------------------------------------------------------------------
# Single-job executor
# ---------------------------------------------------------------------------

def _run_single_job(
    job: PostProcessingJob,
    time_arr: np.ndarray,
    channels: Dict[str, Any],
) -> PostProcessingJobResult:
    """Execute one post-processing job and return its result."""
    t0 = _time_module.monotonic()

    def _fail(
        code: PostProcessingDiagnosticCode, msg: str
    ) -> PostProcessingJobResult:
        return PostProcessingJobResult(
            job_id=job.job_id,
            kind=job.kind,
            success=False,
            diagnostic=code,
            diagnostic_message=msg,
            runtime_seconds=_time_module.monotonic() - t0,
        )

    # --- Signal resolution (for jobs that need signal data) ---
    signal_data: Dict[str, np.ndarray] = {}
    if job.kind in (PostProcessingJobKind.TimeDomain, PostProcessingJobKind.Spectral):
        if not job.signals:
            return _fail(
                PostProcessingDiagnosticCode.InvalidConfiguration,
                f"Job '{job.job_id}': no signals specified",
            )
        for sig_name in job.signals:
            arr, code, msg = _get_signal_data(channels, sig_name)
            if arr is None:
                return _fail(code, f"Job '{job.job_id}': {msg}")
            signal_data[sig_name] = arr

    # --- Window planning ---
    i_start, i_end, code, msg = _resolve_window_indices(time_arr, job.window)
    if code != PostProcessingDiagnosticCode.Ok:
        return _fail(code, f"Job '{job.job_id}': {msg}")

    sample_count = i_end - i_start
    if sample_count < job.window.min_samples:
        return _fail(
            PostProcessingDiagnosticCode.InsufficientSamples,
            f"Job '{job.job_id}': window has {sample_count} samples, "
            f"minimum is {job.window.min_samples}",
        )

    time_window = time_arr[i_start:i_end]

    # --- Time-domain job ---
    if job.kind == PostProcessingJobKind.TimeDomain:
        all_metrics: Dict[str, ScalarMetric] = {}
        all_undefined: List[UndefinedMetricEntry] = []
        multi = len(signal_data) > 1

        for sig_name, sig_arr in signal_data.items():
            if len(sig_arr) != len(time_arr):
                all_undefined.append(UndefinedMetricEntry(
                    sig_name,
                    PostProcessingDiagnosticCode.SamplingMismatch,
                    f"Signal '{sig_name}' length {len(sig_arr)} "
                    f"!= time array length {len(time_arr)}",
                ))
                continue
            x = sig_arr[i_start:i_end]
            m, u = _compute_time_domain_metrics(x, job.metrics, sig_name)
            if multi:
                for k, v in m.items():
                    all_metrics[f"{sig_name}.{k}"] = v
            else:
                all_metrics.update(m)
            all_undefined.extend(u)

        return PostProcessingJobResult(
            job_id=job.job_id,
            kind=job.kind,
            success=True,
            diagnostic=PostProcessingDiagnosticCode.Ok,
            scalar_metrics=all_metrics,
            undefined_metrics=all_undefined,
            signal_names=list(signal_data.keys()),
            window_i_start=i_start,
            window_i_end=i_end,
            sample_count=sample_count,
            runtime_seconds=_time_module.monotonic() - t0,
        )

    # --- Spectral job ---
    if job.kind == PostProcessingJobKind.Spectral:
        sig_name = job.signals[0]
        sig_arr = signal_data[sig_name]

        if len(sig_arr) != len(time_arr):
            return _fail(
                PostProcessingDiagnosticCode.SamplingMismatch,
                f"Job '{job.job_id}': signal '{sig_name}' length {len(sig_arr)} "
                f"!= time array length {len(time_arr)}",
            )

        x = sig_arr[i_start:i_end]
        bins, harmonics, thd, f0, undef = _compute_spectral_metrics(
            x, time_window, job.fundamental_hz, job.n_harmonics, job.window_function
        )

        return PostProcessingJobResult(
            job_id=job.job_id,
            kind=job.kind,
            success=True,
            diagnostic=PostProcessingDiagnosticCode.Ok,
            spectrum_bins=bins,
            harmonics=harmonics,
            thd_pct=thd,
            fundamental_hz=f0,
            undefined_metrics=undef,
            signal_names=[sig_name],
            window_i_start=i_start,
            window_i_end=i_end,
            sample_count=sample_count,
            runtime_seconds=_time_module.monotonic() - t0,
        )

    # --- Power/efficiency job ---
    if job.kind == PostProcessingJobKind.PowerEfficiency:

        def _get_windowed(name: str) -> np.ndarray:
            if not name:
                return np.array([], dtype=np.float64)
            arr, c, m = _get_signal_data(channels, name)
            if arr is None:
                return np.array([], dtype=np.float64)
            if len(arr) != len(time_arr):
                return np.array([], dtype=np.float64)
            return arr[i_start:i_end]

        v_in = _get_windowed(job.input_voltage_signal)
        i_in = _get_windowed(job.input_current_signal)
        v_out = _get_windowed(job.output_voltage_signal)
        i_out = _get_windowed(job.output_current_signal)

        if len(v_in) > 0 and len(i_in) > 0 and len(v_in) != len(i_in):
            return _fail(
                PostProcessingDiagnosticCode.SamplingMismatch,
                f"Job '{job.job_id}': input voltage/current arrays "
                f"have different lengths ({len(v_in)} vs {len(i_in)})",
            )
        if len(v_out) > 0 and len(i_out) > 0 and len(v_out) != len(i_out):
            return _fail(
                PostProcessingDiagnosticCode.SamplingMismatch,
                f"Job '{job.job_id}': output voltage/current arrays "
                f"have different lengths ({len(v_out)} vs {len(i_out)})",
            )

        p_in, p_out, eff, pf, undef = _compute_power_efficiency_metrics(
            v_in, i_in, v_out, i_out
        )

        sig_names = [
            s
            for s in [
                job.input_voltage_signal,
                job.input_current_signal,
                job.output_voltage_signal,
                job.output_current_signal,
            ]
            if s
        ]

        return PostProcessingJobResult(
            job_id=job.job_id,
            kind=job.kind,
            success=True,
            diagnostic=PostProcessingDiagnosticCode.Ok,
            average_input_power=p_in,
            average_output_power=p_out,
            efficiency=eff,
            power_factor=pf,
            undefined_metrics=undef,
            signal_names=sig_names,
            window_i_start=i_start,
            window_i_end=i_end,
            sample_count=sample_count,
            runtime_seconds=_time_module.monotonic() - t0,
        )

    return _fail(
        PostProcessingDiagnosticCode.InvalidConfiguration,
        f"Unknown job kind: {job.kind}",
    )


# ---------------------------------------------------------------------------
# Public entry point
# ---------------------------------------------------------------------------

def run_post_processing(
    result: Any,
    config: PostProcessingOptions,
) -> PostProcessingResult:
    """Execute waveform post-processing jobs on a simulation result.

    Args:
        result: A SimulationResult object with ``.time`` and
                ``.virtual_channels`` attributes (as returned by
                :class:`~pulsim.Simulator`).
        config: :class:`PostProcessingOptions` specifying which jobs to run.

    Returns:
        :class:`PostProcessingResult` containing per-job outputs.

    The function is deterministic: calling it twice with identical inputs
    produces identical outputs (modulo ``runtime_seconds`` telemetry).
    """
    wall_start = _time_module.monotonic()

    time_arr = np.asarray(result.time, dtype=np.float64)
    channels: Dict[str, Any] = dict(result.virtual_channels)

    job_results: List[PostProcessingJobResult] = []
    overall_success = True

    for job in config.jobs:
        jr = _run_single_job(job, time_arr, channels)
        job_results.append(jr)
        if not jr.success:
            overall_success = False

    return PostProcessingResult(
        success=overall_success,
        jobs=job_results,
        total_runtime_seconds=_time_module.monotonic() - wall_start,
        run_count=1,
    )


# ---------------------------------------------------------------------------
# YAML parsing helpers (for simulation.post_processing block)
# ---------------------------------------------------------------------------

_WINDOW_MODE_MAP: Dict[str, PostProcessingWindowMode] = {
    "time": PostProcessingWindowMode.Time,
    "index": PostProcessingWindowMode.Index,
    "cycle": PostProcessingWindowMode.Cycle,
}

_JOB_KIND_MAP: Dict[str, PostProcessingJobKind] = {
    "time_domain": PostProcessingJobKind.TimeDomain,
    "spectral": PostProcessingJobKind.Spectral,
    "power_efficiency": PostProcessingJobKind.PowerEfficiency,
    "power": PostProcessingJobKind.PowerEfficiency,
    "efficiency": PostProcessingJobKind.PowerEfficiency,
}

_WINDOW_FUNC_MAP: Dict[str, WindowFunction] = {
    "rectangular": WindowFunction.Rectangular,
    "rect": WindowFunction.Rectangular,
    "hann": WindowFunction.Hann,
    "hanning": WindowFunction.Hann,
    "hamming": WindowFunction.Hamming,
    "blackman": WindowFunction.Blackman,
    "flattop": WindowFunction.FlatTop,
    "flat_top": WindowFunction.FlatTop,
}


def _parse_window_spec(
    node: Dict[str, Any],
    path: str,
    errors: List[str],
) -> PostProcessingWindowSpec:
    spec = PostProcessingWindowSpec()
    if not isinstance(node, dict):
        errors.append(
            f"[PULSIM_PP_E_WINDOW_INVALID] {path}: window must be a mapping"
        )
        return spec

    mode_raw = str(node.get("mode", "time")).lower()
    mode = _WINDOW_MODE_MAP.get(mode_raw)
    if mode is None:
        accepted = list(_WINDOW_MODE_MAP.keys())
        errors.append(
            f"[PULSIM_PP_E_WINDOW_MODE] {path}.mode: unsupported mode "
            f"'{mode_raw}' (accepted: {accepted})"
        )
    else:
        spec.mode = mode

    if "t_start" in node:
        spec.t_start = float(node["t_start"])
    if "t_end" in node:
        spec.t_end = float(node["t_end"])
    if "i_start" in node:
        spec.i_start = int(node["i_start"])
    if "i_end" in node:
        spec.i_end = int(node["i_end"])
    if "cycle_start" in node:
        spec.cycle_start = int(node["cycle_start"])
    if "cycle_end" in node:
        spec.cycle_end = int(node["cycle_end"])
    if "period" in node:
        spec.period = float(node["period"])
    if "min_samples" in node:
        spec.min_samples = int(node["min_samples"])

    return spec


def _parse_post_processing_job(
    node: Any,
    idx: int,
    path: str,
    errors: List[str],
) -> Optional[PostProcessingJob]:
    if not isinstance(node, dict):
        errors.append(
            f"[PULSIM_PP_E_JOB_INVALID] {path}[{idx}]: job must be a mapping"
        )
        return None

    job_id = str(node.get("id", f"job_{idx}"))
    kind_raw = str(node.get("kind", "")).lower()

    if not kind_raw:
        accepted = list(_JOB_KIND_MAP.keys())
        errors.append(
            f"[PULSIM_PP_E_JOB_KIND] {path}[{idx}].kind: "
            f"missing required field 'kind' (accepted: {accepted})"
        )
        return None

    kind = _JOB_KIND_MAP.get(kind_raw)
    if kind is None:
        accepted = list(_JOB_KIND_MAP.keys())
        errors.append(
            f"[PULSIM_PP_E_JOB_KIND] {path}[{idx}].kind: "
            f"unsupported kind '{kind_raw}' (accepted: {accepted})"
        )
        return None

    job = PostProcessingJob(job_id=job_id, kind=kind)

    # Signals
    signals_raw = node.get("signals", [])
    if isinstance(signals_raw, str):
        signals_raw = [signals_raw]
    job.signals = [str(s) for s in signals_raw]

    # Window
    if "window" in node:
        job.window = _parse_window_spec(
            node["window"], f"{path}[{idx}].window", errors
        )

    # Time-domain metric list
    if "metrics" in node:
        m = node["metrics"]
        if isinstance(m, str):
            job.metrics = [x.strip() for x in m.split(",")]
        elif isinstance(m, list):
            job.metrics = [str(x) for x in m]

    # Spectral options
    if "n_harmonics" in node:
        job.n_harmonics = int(node["n_harmonics"])
    if "fundamental_hz" in node:
        job.fundamental_hz = float(node["fundamental_hz"])
    if "window_function" in node:
        wf_raw = str(node["window_function"]).lower()
        wf = _WINDOW_FUNC_MAP.get(wf_raw)
        if wf is None:
            accepted = list(_WINDOW_FUNC_MAP.keys())
            errors.append(
                f"[PULSIM_PP_E_WINDOW_FUNC] {path}[{idx}].window_function: "
                f"unsupported function '{wf_raw}' (accepted: {accepted})"
            )
        else:
            job.window_function = wf

    # Power/efficiency signal bindings
    for attr, keys in [
        ("input_voltage_signal", ["input_voltage", "v_in", "vin"]),
        ("input_current_signal", ["input_current", "i_in", "iin"]),
        ("output_voltage_signal", ["output_voltage", "v_out", "vout"]),
        ("output_current_signal", ["output_current", "i_out", "iout"]),
    ]:
        for key in keys:
            if key in node:
                setattr(job, attr, str(node[key]))
                break

    return job


def parse_post_processing_yaml(
    node: Any,
    errors: Optional[List[str]] = None,
) -> PostProcessingOptions:
    """Parse a ``simulation.post_processing`` YAML node into PostProcessingOptions.

    Args:
        node: The raw Python dict from a parsed YAML ``simulation.post_processing``
              block.  Pass ``None`` for no post-processing (returns empty options).
        errors: Optional list; parse errors are appended here using stable
                machine-readable ``[PULSIM_PP_E_*]`` prefixes.

    Returns:
        :class:`PostProcessingOptions` (empty if *node* is ``None`` or invalid).

    Example YAML block::

        simulation:
          post_processing:
            jobs:
              - id: output_metrics
                kind: time_domain
                signals: ["V(out)"]
                window: {mode: time, t_start: 1.5e-3, t_end: 2.0e-3}
                metrics: [rms, mean, min, max, p2p, ripple]
              - id: input_fft
                kind: spectral
                signals: ["I(L1)"]
                window: {mode: time, t_start: 1.5e-3, t_end: 2.0e-3}
                n_harmonics: 5
                window_function: hann
              - id: efficiency
                kind: power_efficiency
                window: {mode: time, t_start: 1.5e-3, t_end: 2.0e-3}
                input_voltage: "V(Vdc)"
                input_current: "I(Vdc)"
                output_voltage: "V(out)"
                output_current: "I(Rload)"
    """
    if errors is None:
        errors = []

    opts = PostProcessingOptions()

    if node is None:
        return opts

    if not isinstance(node, dict):
        errors.append(
            "[PULSIM_PP_E_BLOCK_INVALID] simulation.post_processing: "
            "must be a mapping (dict)"
        )
        return opts

    jobs_raw = node.get("jobs", [])
    if not isinstance(jobs_raw, list):
        errors.append(
            "[PULSIM_PP_E_JOBS_INVALID] simulation.post_processing.jobs: "
            "must be a list"
        )
        return opts

    path = "simulation.post_processing.jobs"
    for idx, job_node in enumerate(jobs_raw):
        job = _parse_post_processing_job(job_node, idx, path, errors)
        if job is not None:
            opts.jobs.append(job)

    return opts



"""
Basic Component Validation Tests (Level 1)

Validates fundamental circuit components against analytical solutions:
- Resistor: Ohm's law V = IR
- Capacitor: I = C * dV/dt
- Inductor: V = L * dI/dt
- Voltage Source: Constant voltage output
- Current Source: Constant current output
"""

import pytest
import numpy as np
import pulsim as sl



# =============================================================================
# Test: Resistor - Ohm's Law
# =============================================================================

class TestResistor:
    """Validate resistor behavior (Ohm's law: V = IR)."""

    @pytest.mark.parametrize("R,V", [
        (100.0, 10.0),
        (1000.0, 5.0),
        (10.0, 1.0),
        (1e6, 100.0),  # High resistance
        (0.1, 1.0),    # Low resistance
    ])
    def test_resistor_ohms_law(self, R, V):
        """Verify V = IR for various resistor values."""
        circuit = sl.Circuit()
        circuit.add_voltage_source("V1", "in", "0", V)
        circuit.add_resistor("R1", "in", "0", R)

        opts = sl.SimulationOptions()
        opts.tstart = 0.0
        opts.tstop = 1e-6  # Very short - just need DC operating point
        opts.dt = 1e-7

        sim = sl.Simulator(circuit, opts)
        result = sim.run_transient()
        assert result.final_status == sl.SolverStatus.Success

        # Get current through voltage source (which equals current through R)
        signal_names = result.signal_names
        data_matrix = np.array(result.data)
        signal_data = {name: data_matrix[:, i] for i, name in enumerate(signal_names)}

        # Current through voltage source
        I_measured = signal_data["I(V1)"][-1]
        I_expected = V / R

        # Current flows into positive terminal of voltage source
        relative_error = abs(abs(I_measured) - I_expected) / I_expected
        print(f"\nR={R}Ω, V={V}V: I_measured={I_measured:.6e}A, I_expected={I_expected:.6e}A")
        print(f"Relative error: {relative_error:.6e}")

        assert relative_error < 1e-6, f"Ohm's law violation: {relative_error:.2e}"

    def test_resistor_voltage_divider(self):
        """Verify voltage divider formula: Vout = Vin * R2 / (R1 + R2)."""
        Vin = 10.0
        R1 = 1000.0
        R2 = 1000.0
        Vout_expected = Vin * R2 / (R1 + R2)

        circuit = sl.Circuit()
        circuit.add_voltage_source("V1", "in", "0", Vin)
        circuit.add_resistor("R1", "in", "out", R1)
        circuit.add_resistor("R2", "out", "0", R2)

        opts = sl.SimulationOptions()
        opts.tstart = 0.0
        opts.tstop = 1e-6
        opts.dt = 1e-7

        sim = sl.Simulator(circuit, opts)
        result = sim.run_transient()
        assert result.final_status == sl.SolverStatus.Success

        signal_names = result.signal_names
        data_matrix = np.array(result.data)
        signal_data = {name: data_matrix[:, i] for i, name in enumerate(signal_names)}

        Vout_measured = signal_data["V(out)"][-1]
        relative_error = abs(Vout_measured - Vout_expected) / Vout_expected

        print(f"\nVoltage divider: Vout={Vout_measured:.6f}V (expected {Vout_expected:.6f}V)")
        print(f"Relative error: {relative_error:.6e}")

        assert relative_error < 1e-6, f"Voltage divider error: {relative_error:.2e}"

    @pytest.mark.parametrize("R1,R2,R3", [
        (100.0, 200.0, 300.0),
        (1000.0, 1000.0, 1000.0),
    ])
    def test_resistor_series_parallel(self, R1, R2, R3):
        """Verify series and parallel resistor combinations."""
        V = 10.0

        # R1 in series with (R2 || R3)
        R_parallel = (R2 * R3) / (R2 + R3)
        R_total = R1 + R_parallel
        I_expected = V / R_total

        circuit = sl.Circuit()
        circuit.add_voltage_source("V1", "in", "0", V)
        circuit.add_resistor("R1", "in", "mid", R1)
        circuit.add_resistor("R2", "mid", "0", R2)
        circuit.add_resistor("R3", "mid", "0", R3)

        opts = sl.SimulationOptions()
        opts.tstart = 0.0
        opts.tstop = 1e-6
        opts.dt = 1e-7

        sim = sl.Simulator(circuit, opts)
        result = sim.run_transient()
        assert result.final_status == sl.SolverStatus.Success

        signal_names = result.signal_names
        data_matrix = np.array(result.data)
        signal_data = {name: data_matrix[:, i] for i, name in enumerate(signal_names)}

        I_measured = abs(signal_data["I(V1)"][-1])
        relative_error = abs(I_measured - I_expected) / I_expected

        print(f"\nSeries-parallel: I={I_measured:.6e}A (expected {I_expected:.6e}A)")
        print(f"Relative error: {relative_error:.6e}")

        assert relative_error < 1e-6, f"Series-parallel error: {relative_error:.2e}"


# =============================================================================
# Test: Capacitor - I = C * dV/dt
# =============================================================================

class TestCapacitor:
    """Validate capacitor behavior (I = C * dV/dt)."""

    def test_capacitor_charging_current(self):
        """Verify initial charging current I = V/R for RC circuit."""
        V = 10.0
        R = 1000.0
        C = 1e-6
        I_initial_expected = V / R  # At t=0, capacitor is short circuit

        circuit = sl.Circuit()
        circuit.add_voltage_source("V1", "in", "0", V)
        circuit.add_resistor("R1", "in", "out", R)
        circuit.add_capacitor("C1", "out", "0", C, ic=0.0)

        opts = sl.SimulationOptions()
        opts.tstart = 0.0
        opts.tstop = 1e-6  # Very short time
        opts.dt = 1e-9
        opts.use_ic = True

        sim = sl.Simulator(circuit, opts)
        result = sim.run_transient()
        assert result.final_status == sl.SolverStatus.Success

        signal_names = result.signal_names
        data_matrix = np.array(result.data)
        signal_data = {name: data_matrix[:, i] for i, name in enumerate(signal_names)}

        # Current at very beginning
        I_initial = abs(signal_data["I(V1)"][1])  # Skip t=0
        relative_error = abs(I_initial - I_initial_expected) / I_initial_expected

        print(f"\nInitial charging current: {I_initial:.6e}A (expected {I_initial_expected:.6e}A)")
        print(f"Relative error: {relative_error:.4%}")

        # Allow some tolerance due to numerical integration
        assert relative_error < 0.01, f"Initial current error: {relative_error:.2%}"

    def test_capacitor_charge_conservation(self):
        """Verify Q = CV relationship."""
        V = 5.0
        C = 10e-6
        C * V

        circuit = sl.Circuit()
        circuit.add_voltage_source("V1", "in", "0", V)
        circuit.add_capacitor("C1", "in", "0", C, ic=0.0)

        # Small resistor to allow current flow (pure V-C would be singular)
        circuit.add_resistor("R1", "in", "0", 0.001)

        opts = sl.SimulationOptions()
        opts.tstart = 0.0
        opts.tstop = 1e-3  # Allow time to charge
        opts.dt = 1e-6
        opts.use_ic = True

        sim = sl.Simulator(circuit, opts)
        result = sim.run_transient()
        assert result.final_status == sl.SolverStatus.Success

        signal_names = result.signal_names
        data_matrix = np.array(result.data)
        signal_data = {name: data_matrix[:, i] for i, name in enumerate(signal_names)}

        # Final voltage across capacitor should be V
        V_cap_final = signal_data["V(in)"][-1]

        print(f"\nCapacitor voltage: {V_cap_final:.6f}V (expected {V:.6f}V)")
        assert abs(V_cap_final - V) / V < 1e-4, "Capacitor did not charge to source voltage"

    def test_capacitor_parallel(self):
        """Verify parallel capacitors: C_total = C1 + C2."""
        V = 10.0
        R = 1000.0
        C1 = 1e-6
        C2 = 2e-6
        C_total = C1 + C2
        tau_expected = R * C_total

        circuit = sl.Circuit()
        circuit.add_voltage_source("V1", "in", "0", V)
        circuit.add_resistor("R1", "in", "out", R)
        circuit.add_capacitor("C1", "out", "0", C1, ic=0.0)
        circuit.add_capacitor("C2", "out", "0", C2, ic=0.0)

        opts = sl.SimulationOptions()
        opts.tstart = 0.0
        opts.tstop = 5 * tau_expected
        opts.dt = tau_expected / 500
        opts.use_ic = True
        opts.dtmax = opts.dt

        sim = sl.Simulator(circuit, opts)
        result = sim.run_transient()
        assert result.final_status == sl.SolverStatus.Success

        time = np.array(result.time)
        signal_names = result.signal_names
        data_matrix = np.array(result.data)
        signal_data = {name: data_matrix[:, i] for i, name in enumerate(signal_names)}

        v_out = signal_data["V(out)"]

        # At t = tau, V should be V * (1 - 1/e) ≈ 0.632 * V
        idx_tau = np.argmin(np.abs(time - tau_expected))
        v_at_tau = v_out[idx_tau]
        v_expected_at_tau = V * (1 - np.exp(-1))

        relative_error = abs(v_at_tau - v_expected_at_tau) / v_expected_at_tau
        print(f"\nParallel caps: V(τ)={v_at_tau:.4f}V (expected {v_expected_at_tau:.4f}V)")
        print(f"Relative error: {relative_error:.4%}")

        assert relative_error < 0.01, f"Parallel capacitor error: {relative_error:.2%}"


# =============================================================================
# Test: Inductor - V = L * dI/dt
# =============================================================================

class TestInductor:
    """Validate inductor behavior (V = L * dI/dt)."""

    def test_inductor_initial_current_zero(self):
        """Verify inductor blocks sudden current change (acts as open circuit initially)."""
        V = 10.0
        R = 100.0
        L = 10e-3

        circuit = sl.Circuit()
        circuit.add_voltage_source("V1", "in", "0", V)
        circuit.add_resistor("R1", "in", "out", R)
        circuit.add_inductor("L1", "out", "0", L, ic=0.0)

        opts = sl.SimulationOptions()
        opts.tstart = 0.0
        opts.tstop = 1e-7  # Very short time
        opts.dt = 1e-9
        opts.use_ic = True

        sim = sl.Simulator(circuit, opts)
        result = sim.run_transient()
        assert result.final_status == sl.SolverStatus.Success

        signal_names = result.signal_names
        data_matrix = np.array(result.data)
        signal_data = {name: data_matrix[:, i] for i, name in enumerate(signal_names)}

        # Current should start at 0
        I_initial = signal_data["I(L1)"][0]
        print(f"\nInitial inductor current: {I_initial:.6e}A (expected ~0)")

        assert abs(I_initial) < 1e-6, f"Initial current not zero: {I_initial}"

    def test_inductor_final_current(self):
        """Verify inductor acts as short circuit at DC steady state."""
        V = 10.0
        R = 100.0
        L = 10e-3
        tau = L / R
        I_final_expected = V / R

        circuit = sl.Circuit()
        circuit.add_voltage_source("V1", "in", "0", V)
        circuit.add_resistor("R1", "in", "out", R)
        circuit.add_inductor("L1", "out", "0", L, ic=0.0)

        opts = sl.SimulationOptions()
        opts.tstart = 0.0
        opts.tstop = 10 * tau  # Long enough for steady state
        opts.dt = tau / 100
        opts.use_ic = True

        sim = sl.Simulator(circuit, opts)
        result = sim.run_transient()
        assert result.final_status == sl.SolverStatus.Success

        signal_names = result.signal_names
        data_matrix = np.array(result.data)
        signal_data = {name: data_matrix[:, i] for i, name in enumerate(signal_names)}

        I_final = signal_data["I(L1)"][-1]
        relative_error = abs(I_final - I_final_expected) / I_final_expected

        print(f"\nFinal inductor current: {I_final:.6f}A (expected {I_final_expected:.6f}A)")
        print(f"Relative error: {relative_error:.4%}")

        assert relative_error < 0.01, f"Final current error: {relative_error:.2%}"

    def test_inductor_series(self):
        """Verify series inductors: L_total = L1 + L2."""
        V = 10.0
        R = 100.0
        L1 = 5e-3
        L2 = 10e-3
        L_total = L1 + L2
        tau_expected = L_total / R

        circuit = sl.Circuit()
        circuit.add_voltage_source("V1", "in", "0", V)
        circuit.add_resistor("R1", "in", "n1", R)
        circuit.add_inductor("L1", "n1", "n2", L1, ic=0.0)
        circuit.add_inductor("L2", "n2", "0", L2, ic=0.0)

        opts = sl.SimulationOptions()
        opts.tstart = 0.0
        opts.tstop = 5 * tau_expected
        opts.dt = tau_expected / 500
        opts.use_ic = True
        opts.dtmax = opts.dt

        sim = sl.Simulator(circuit, opts)
        result = sim.run_transient()
        assert result.final_status == sl.SolverStatus.Success

        time = np.array(result.time)
        signal_names = result.signal_names
        data_matrix = np.array(result.data)
        signal_data = {name: data_matrix[:, i] for i, name in enumerate(signal_names)}

        i_L = signal_data["I(L1)"]  # Same current through both

        # At t = tau, I should be I_final * (1 - 1/e)
        I_final = V / R
        idx_tau = np.argmin(np.abs(time - tau_expected))
        i_at_tau = i_L[idx_tau]
        i_expected_at_tau = I_final * (1 - np.exp(-1))

        relative_error = abs(i_at_tau - i_expected_at_tau) / i_expected_at_tau
        print(f"\nSeries inductors: I(τ)={i_at_tau:.6f}A (expected {i_expected_at_tau:.6f}A)")
        print(f"Relative error: {relative_error:.4%}")

        assert relative_error < 0.01, f"Series inductor error: {relative_error:.2%}"


# =============================================================================
# Test: Voltage Source
# =============================================================================

class TestVoltageSource:
    """Validate voltage source behavior."""

    @pytest.mark.parametrize("V", [1.0, 5.0, 10.0, 100.0, -5.0])
    def test_voltage_source_dc(self, V):
        """Verify DC voltage source maintains constant voltage."""
        circuit = sl.Circuit()
        circuit.add_voltage_source("V1", "out", "0", V)
        circuit.add_resistor("R1", "out", "0", 1000.0)

        opts = sl.SimulationOptions()
        opts.tstart = 0.0
        opts.tstop = 1e-3
        opts.dt = 1e-5

        sim = sl.Simulator(circuit, opts)
        result = sim.run_transient()
        assert result.final_status == sl.SolverStatus.Success

        signal_names = result.signal_names
        data_matrix = np.array(result.data)
        signal_data = {name: data_matrix[:, i] for i, name in enumerate(signal_names)}

        v_out = signal_data["V(out)"]

        # All values should be V
        max_deviation = np.max(np.abs(v_out - V))
        print(f"\nDC voltage source {V}V: max deviation = {max_deviation:.2e}V")

        assert max_deviation < 1e-10, f"Voltage source deviation: {max_deviation}"


# =============================================================================
# Test: Energy Conservation
# =============================================================================

class TestEnergyConservation:
    """Test energy conservation in reactive circuits."""

    def test_lc_oscillator_energy(self):
        """Verify energy conservation in LC oscillator."""
        L = 1e-3
        C = 1e-6
        V0 = 10.0  # Initial capacitor voltage

        # Energy = 0.5 * C * V^2 = 0.5 * L * I^2
        E_initial = 0.5 * C * V0**2
        omega = 1.0 / np.sqrt(L * C)
        period = 2 * np.pi / omega

        circuit = sl.Circuit()
        circuit.add_capacitor("C1", "out", "0", C, ic=V0)
        circuit.add_inductor("L1", "out", "0", L, ic=0.0)
        # Very large resistance for numerical stability, approximating ideal LC
        circuit.add_resistor("R1", "out", "0", 1e12)

        opts = sl.SimulationOptions()
        opts.tstart = 0.0
        opts.tstop = 5 * period
        opts.dt = period / 500
        opts.dtmin = opts.dt / 10  # Avoid edge case issues
        opts.use_ic = True
        # Use BDF2/GEAR2 for excellent energy conservation (~99.97%)
        opts.integration_method = sl.IntegrationMethod.GEAR2
        sim = sl.Simulator(circuit, opts)
        result = sim.run_transient()
        assert result.final_status == sl.SolverStatus.Success

        signal_names = result.signal_names
        data_matrix = np.array(result.data)
        signal_data = {name: data_matrix[:, i] for i, name in enumerate(signal_names)}

        v_C = signal_data["V(out)"]
        i_L = signal_data["I(L1)"]

        # Calculate total energy at each point
        E_C = 0.5 * C * v_C**2
        E_L = 0.5 * L * i_L**2
        E_total = E_C + E_L

        # Energy should be approximately constant (with some numerical damping)
        E_variation = (np.max(E_total) - np.min(E_total)) / E_initial

        print("\nLC oscillator energy conservation:")
        print(f"  Initial energy: {E_initial:.6e} J")
        print(f"  Final energy: {E_total[-1]:.6e} J")
        print(f"  Energy variation: {E_variation*100:.2f}%")

        # Allow some energy loss due to numerical damping
        assert E_variation < 0.2, f"Energy variation too high: {E_variation*100:.1f}%"


# =============================================================================
# Run validation suite
# =============================================================================

def run_component_validation_suite(verbose: bool = True):
    """Run all component validation tests."""
    import subprocess
    result = subprocess.run(
        ["python", "-m", "pytest", __file__, "-v", "--tb=short"],
        capture_output=True,
        text=True,
    )
    print(result.stdout)
    if result.returncode != 0:
        print(result.stderr)
    return result.returncode


if __name__ == "__main__":
    # Run a quick manual test
    print("Testing basic components...")

    # Resistor
    print("\n=== Resistor Test ===")
    test = TestResistor()
    test.test_resistor_ohms_law(100.0, 10.0)
    test.test_resistor_voltage_divider()

    # Capacitor
    print("\n=== Capacitor Test ===")
    test = TestCapacitor()
    test.test_capacitor_charging_current()

    # Inductor
    print("\n=== Inductor Test ===")
    test = TestInductor()
    test.test_inductor_final_current()

    # Energy
    print("\n=== Energy Conservation Test ===")
    test = TestEnergyConservation()
    test.test_lc_oscillator_energy()

    print("\n\nAll component tests passed!")



"""Base classes for validation framework."""

from dataclasses import dataclass
from typing import Optional, Callable, Any, Dict, List
from enum import Enum
import numpy as np
import time


class ValidationLevel(Enum):
    """Validation complexity levels."""
    LINEAR = 1         # Linear circuits with analytical solution
    DC_ANALYSIS = 2    # DC analysis
    NONLINEAR = 3      # Nonlinear components
    CONVERTER = 4      # Power converters


@dataclass
class ValidationResult:
    """Result of a validation test."""
    test_name: str
    passed: bool
    pulsim_times: np.ndarray
    pulsim_values: np.ndarray
    reference_times: np.ndarray
    reference_values: np.ndarray
    max_error: float
    rms_error: float
    max_relative_error: float
    tolerance: float
    execution_time_ms: float
    notes: str = ""
    max_error_threshold: Optional[float] = None
    rms_error_threshold: Optional[float] = None

    def summary(self) -> str:
        """Return a summary string of the validation result."""
        status = "PASS" if self.passed else "FAIL"
        return (
            f"{self.test_name}: {status}\n"
            f"  Max Error: {self.max_error:.6e}\n"
            f"  RMS Error: {self.rms_error:.6e}\n"
            f"  Max Rel Error: {self.max_relative_error*100:.4f}%\n"
            f"  Tolerance: {self.tolerance*100:.2f}%\n"
            f"  Exec Time: {self.execution_time_ms:.2f} ms"
        )


@dataclass
class CircuitDefinition:
    """Definition of a test circuit."""
    name: str
    description: str
    level: ValidationLevel
    build_circuit: Optional[Callable[[], Any]] = None  # Returns pulsim.Circuit
    pulsim_builder: Optional[Callable[[], Any]] = None
    analytical_solution: Optional[Callable[[np.ndarray], np.ndarray]] = None
    t_start: float = 0.0
    t_stop: float = 1e-3
    dt: float = 1e-6
    node_index: int = 1  # Node index to compare (0 is usually internal)
    tolerance: float = 0.01  # 1% default
    dc_tolerance: float = 0.0001  # 0.01% for DC
    spice_netlist: Optional[str] = None
    use_zero_ic: bool = False  # If True, use IC=0 instead of DC operating point
    custom_ic: Optional[Callable[[Any], np.ndarray]] = None  # Custom IC function(circuit) -> x0
    # New-style validation fields (level2+)
    tstart: Optional[float] = None
    tstop: Optional[float] = None
    compare_nodes: Optional[Dict[str, str]] = None
    circuit_params: Optional[Dict[str, float]] = None
    pulsim_options: Optional[Dict[str, Any]] = None
    max_error_tolerance: Optional[float] = None
    rms_error_tolerance: Optional[float] = None

    def __post_init__(self):
        if self.pulsim_builder is None:
            self.pulsim_builder = self.build_circuit
        if self.tstart is not None:
            self.t_start = self.tstart
        if self.tstop is not None:
            self.t_stop = self.tstop


class ValidationTest:
    """Base class for validation tests."""

    def __init__(self, circuit_def: CircuitDefinition):
        self.circuit_def = circuit_def
        self._import_pulsim()

    def _import_pulsim(self):
        """Import pulsim module."""
        import pulsim as ps
        self.ps = ps

    def _build_circuit(self):
        builder = self.circuit_def.pulsim_builder or self.circuit_def.build_circuit
        if builder is None:
            raise ValueError("CircuitDefinition missing build_circuit/pulsim_builder")
        return builder()

    def run_pulsim_dc(self):
        """Run DC analysis in Pulsim."""
        circuit = self._build_circuit()
        return self.ps.dc_operating_point(circuit)

    def run_pulsim_transient(self) -> tuple:
        """Run transient simulation in Pulsim.

        Returns:
            tuple: (times, states, success, message)
        """
        circuit = self._build_circuit()

        # Determine initial conditions
        if self.circuit_def.custom_ic is not None:
            # Use custom IC function
            x0 = self.circuit_def.custom_ic(circuit)
        elif self.circuit_def.use_zero_ic:
            # Use zero initial conditions (for step response tests)
            x0 = np.zeros(circuit.system_size())
            # Set voltage source nodes to their DC values
            dc_result = self.ps.dc_operating_point(circuit)
            if dc_result.success:
                # Copy only the voltage source node values (usually node 0)
                # Keep capacitor/inductor nodes at 0
                x0[0] = dc_result.newton_result.solution[0]
        else:
            # DC initial condition (steady state)
            dc_result = self.ps.dc_operating_point(circuit)
            if not dc_result.success:
                raise RuntimeError(f"DC failed: {dc_result.message}")
            x0 = dc_result.newton_result.solution

        # Transient
        times, states, success, msg = self.ps.run_transient(
            circuit,
            self.circuit_def.t_start,
            self.circuit_def.t_stop,
            self.circuit_def.dt,
            x0
        )

        if not success:
            raise RuntimeError(f"Transient failed: {msg}")

        return np.array(times), states, success, msg

    def validate(self, use_analytical: bool = True) -> List[ValidationResult]:
        """Run validation for the configured circuit definition."""
        # New-style validation with compare_nodes
        if self.circuit_def.compare_nodes:
            circuit = self._build_circuit()
            opts = self.ps.SimulationOptions()
            opts.tstart = self.circuit_def.t_start
            opts.tstop = self.circuit_def.t_stop
            opts.dt = self.circuit_def.dt

            for key, value in (self.circuit_def.pulsim_options or {}).items():
                setattr(opts, key, value)

            sim = self.ps.Simulator(circuit, opts)
            start = time.perf_counter()
            result = sim.run_transient()
            exec_time = (time.perf_counter() - start) * 1000

            if result.final_status != self.ps.SolverStatus.Success:
                max_tol = self.circuit_def.max_error_tolerance or self.circuit_def.tolerance
                rms_tol = self.circuit_def.rms_error_tolerance or self.circuit_def.tolerance
                return [ValidationResult(
                    test_name=self.circuit_def.name,
                    passed=False,
                    pulsim_times=np.array([0.0]),
                    pulsim_values=np.array([0.0]),
                    reference_times=np.array([0.0]),
                    reference_values=np.array([0.0]),
                    max_error=float("inf"),
                    rms_error=float("inf"),
                    max_relative_error=float("inf"),
                    tolerance=max_tol,
                    execution_time_ms=exec_time,
                    notes=result.message,
                    max_error_threshold=max_tol,
                    rms_error_threshold=rms_tol,
                )]

            times = np.array(result.time)
            data_matrix = np.array(result.data)
            signal_index = {name: i for i, name in enumerate(result.signal_names)}

            results: List[ValidationResult] = []
            for pulsim_signal, analytical_key in self.circuit_def.compare_nodes.items():
                if pulsim_signal not in signal_index:
                    raise ValueError(f"Signal '{pulsim_signal}' not found in results")

                pulsim_values = data_matrix[:, signal_index[pulsim_signal]]

                ref_values = None
                if use_analytical and self.circuit_def.analytical_solution:
                    params = self.circuit_def.circuit_params or {}
                    ref_values = self.circuit_def.analytical_solution(times, analytical_key, params)

                if ref_values is None:
                    raise ValueError("No analytical solution defined for this circuit")

                errors = np.abs(pulsim_values - ref_values)
                max_error = float(np.max(errors))
                rms_error = float(np.sqrt(np.mean(errors**2)))
                max_ref = float(np.max(np.abs(ref_values)))
                max_rel_error = max_error / max_ref if max_ref > 1e-12 else max_error

                max_tol = self.circuit_def.max_error_tolerance or self.circuit_def.tolerance
                rms_tol = self.circuit_def.rms_error_tolerance or self.circuit_def.tolerance
                passed = (max_error <= max_tol) and (rms_error <= rms_tol)

                test_name = self.circuit_def.name
                if len(self.circuit_def.compare_nodes) > 1:
                    test_name = f"{test_name}:{pulsim_signal}"

                results.append(ValidationResult(
                    test_name=test_name,
                    passed=passed,
                    pulsim_times=times,
                    pulsim_values=pulsim_values,
                    reference_times=times,
                    reference_values=ref_values,
                    max_error=max_error,
                    rms_error=rms_error,
                    max_relative_error=max_rel_error,
                    tolerance=max_tol,
                    execution_time_ms=exec_time,
                    max_error_threshold=max_tol,
                    rms_error_threshold=rms_tol,
                ))

            return results

        # Fallback to legacy transient validation
        return [self.validate_transient()]

    def run_analytical(self, times: np.ndarray) -> Optional[np.ndarray]:
        """Calculate analytical solution."""
        if self.circuit_def.analytical_solution:
            return self.circuit_def.analytical_solution(times)
        return None

    def validate_transient(self) -> ValidationResult:
        """Validate transient simulation against analytical solution."""
        start = time.perf_counter()
        times, states, success, msg = self.run_pulsim_transient()
        exec_time = (time.perf_counter() - start) * 1000

        # Extract node of interest
        node_idx = self.circuit_def.node_index
        pulsim_values = np.array([s[node_idx] for s in states])

        # Analytical solution
        ref_values = self.run_analytical(times)
        if ref_values is None:
            raise ValueError("No analytical solution defined for this circuit")

        result = self._compare(times, pulsim_values, times, ref_values)
        result.execution_time_ms = exec_time
        return result

    def validate_dc(self, expected_values: dict) -> ValidationResult:
        """Validate DC analysis against expected values.

        Args:
            expected_values: Dict mapping node index to expected voltage
        """
        start = time.perf_counter()
        dc_result = self.run_pulsim_dc()
        exec_time = (time.perf_counter() - start) * 1000

        if not dc_result.success:
            return ValidationResult(
                test_name=self.circuit_def.name,
                passed=False,
                pulsim_times=np.array([0]),
                pulsim_values=np.array([0]),
                reference_times=np.array([0]),
                reference_values=np.array([0]),
                max_error=float('inf'),
                rms_error=float('inf'),
                max_relative_error=float('inf'),
                tolerance=self.circuit_def.dc_tolerance,
                execution_time_ms=exec_time,
                notes=f"DC analysis failed: {dc_result.message}"
            )

        solution = dc_result.newton_result.solution

        errors = []
        for node_idx, expected in expected_values.items():
            if node_idx < len(solution):
                actual = solution[node_idx]
                error = abs(actual - expected)
                rel_error = error / abs(expected) if expected != 0 else error
                errors.append((error, rel_error))

        max_error = max(e[0] for e in errors) if errors else 0
        max_rel_error = max(e[1] for e in errors) if errors else 0
        rms_error = np.sqrt(np.mean([e[0]**2 for e in errors])) if errors else 0

        passed = max_rel_error <= self.circuit_def.dc_tolerance

        return ValidationResult(
            test_name=self.circuit_def.name,
            passed=passed,
            pulsim_times=np.array([0]),
            pulsim_values=np.array(list(solution)),
            reference_times=np.array([0]),
            reference_values=np.array([expected_values.get(i, 0) for i in range(len(solution))]),
            max_error=max_error,
            rms_error=rms_error,
            max_relative_error=max_rel_error,
            tolerance=self.circuit_def.dc_tolerance,
            execution_time_ms=exec_time
        )

    def _compare(self, pulsim_times, pulsim_values, ref_times, ref_values) -> ValidationResult:
        """Compare Pulsim results vs reference."""
        from .comparator import compare_results

        return compare_results(
            test_name=self.circuit_def.name,
            pulsim_times=pulsim_times,
            pulsim_values=pulsim_values,
            ref_times=ref_times,
            ref_values=ref_values,
            tolerance=self.circuit_def.tolerance
        )



"""Type stubs for pulsim._pulsim C++ extension (pybind11).

This file exists so that static analysis tools (griffe, mypy, pyright)
can resolve imports from this compiled extension without needing the
actual binary to be built.
"""

from typing import Dict, List, Optional, Tuple
from enum import Enum

# =============================================================================
# Enums
# =============================================================================

class DeviceType(Enum):
    Resistor = ...
    Capacitor = ...
    Inductor = ...
    VoltageSource = ...
    CurrentSource = ...
    Diode = ...
    Switch = ...
    MOSFET = ...
    IGBT = ...
    Transformer = ...

class SolverStatus(Enum):
    Success = ...
    MaxIterationsReached = ...
    SingularMatrix = ...
    NumericalError = ...
    ConvergenceStall = ...
    Diverging = ...

class DCStrategy(Enum):
    Direct = ...
    GminStepping = ...
    SourceStepping = ...
    PseudoTransient = ...
    Auto = ...

class RLCDamping(Enum):
    Underdamped = ...
    Critical = ...
    Overdamped = ...

class DeviceHint(Enum):
    None_ = ...
    DiodeAnode = ...
    DiodeCathode = ...
    MOSFETGate = ...
    MOSFETDrain = ...
    MOSFETSource = ...
    BJTBase = ...
    BJTCollector = ...
    BJTEmitter = ...
    SupplyPositive = ...
    SupplyNegative = ...
    Ground = ...

class SIMDLevel(Enum):
    None_ = ...
    SSE2 = ...
    SSE4 = ...
    AVX = ...
    AVX2 = ...
    AVX512 = ...
    NEON = ...

class Integrator(Enum):
    Trapezoidal = ...
    BDF1 = ...
    BDF2 = ...
    BDF3 = ...
    BDF4 = ...
    BDF5 = ...
    Gear = ...
    TRBDF2 = ...
    RosenbrockW = ...
    SDIRK2 = ...

class TimestepMethod(Enum):
    StepDoubling = ...
    Richardson = ...

class StepMode(Enum):
    Fixed = ...
    Variable = ...

class ControlUpdateMode(Enum):
    Auto = ...
    Continuous = ...
    Discrete = ...

class FormulationMode(Enum):
    ProjectedWrapper = ...
    Direct = ...

class FrequencyAnalysisMode(Enum):
    OpenLoopTransfer = ...
    ClosedLoopTransfer = ...
    InputImpedance = ...
    OutputImpedance = ...

class FrequencyAnchorMode(Enum):
    DC = ...
    Periodic = ...
    Averaged = ...
    Auto = ...

class FrequencySweepScale(Enum):
    Logarithmic = ...
    Linear = ...

class FrequencyMetricUndefinedReason(Enum):
    None_ = ...
    NotTransferMode = ...
    NoGainCrossover = ...
    NoPhaseCrossover = ...

class AveragedConverterTopology(Enum):
    Buck = ...
    Boost = ...
    BuckBoost = ...

class AveragedOperatingMode(Enum):
    CCM = ...
    DCM = ...
    Auto = ...

class AveragedEnvelopePolicy(Enum):
    Strict = ...
    Warn = ...

# =============================================================================
# Device Classes
# =============================================================================

class Circuit:
    def __init__(self) -> None: ...
    def add_node(self, name: str) -> int: ...
    def get_node(self, name: str) -> int: ...
    @staticmethod
    def ground() -> int: ...
    def num_nodes(self) -> int: ...
    def num_branches(self) -> int: ...
    def system_size(self) -> int: ...
    def node_name(self, index: int) -> str: ...
    def node_names(self) -> List[str]: ...
    def signal_names(self) -> List[str]: ...
    def initial_state(self) -> List[float]: ...
    def num_devices(self) -> int: ...
    def add_virtual_component(
        self,
        type: str,
        name: str,
        nodes: List[int],
        numeric_params: Dict[str, float] = ...,
        metadata: Dict[str, str] = ...,
    ) -> None: ...
    def num_virtual_components(self) -> int: ...
    def virtual_components(self) -> List["VirtualComponent"]: ...
    def virtual_component_names(self) -> List[str]: ...
    def virtual_channel_metadata(self) -> Dict[str, "VirtualChannelMetadata"]: ...
    @staticmethod
    def mixed_domain_phase_order() -> List[str]: ...
    def execute_mixed_domain_step(self, x: List[float], time: float) -> "MixedDomainStepResult": ...
    def evaluate_virtual_signals(self, x: List[float]) -> Dict[str, float]: ...
    def set_timestep(self, dt: float) -> None: ...
    def timestep(self) -> float: ...

class VirtualComponent:
    type: str
    name: str
    nodes: List[int]
    numeric_params: Dict[str, float]
    metadata: Dict[str, str]
    def __init__(self) -> None: ...

class MixedDomainStepResult:
    phase_order: List[str]
    channel_values: Dict[str, float]
    def __init__(self) -> None: ...

class VirtualChannelMetadata:
    component_type: str
    component_name: str
    source_component: str
    domain: str
    unit: str
    nodes: List[int]
    def __init__(self) -> None: ...

class Resistor:
    def __init__(self, resistance: float, name: str = "") -> None: ...
    def resistance(self) -> float: ...
    def name(self) -> str: ...

class Capacitor:
    def __init__(self, capacitance: float, initial_voltage: float = 0.0, name: str = "") -> None: ...
    def capacitance(self) -> float: ...
    def name(self) -> str: ...
    def set_timestep(self, dt: float) -> None: ...

class Inductor:
    def __init__(self, inductance: float, initial_current: float = 0.0, name: str = "") -> None: ...
    def inductance(self) -> float: ...
    def name(self) -> str: ...
    def set_timestep(self, dt: float) -> None: ...

class VoltageSource:
    def __init__(self, voltage: float, name: str = "") -> None: ...
    def voltage(self) -> float: ...
    def name(self) -> str: ...

class CurrentSource:
    def __init__(self, current: float, name: str = "") -> None: ...
    def current(self) -> float: ...
    def name(self) -> str: ...

class IdealDiode:
    def __init__(self, name: str = "") -> None: ...
    def name(self) -> str: ...

class IdealSwitch:
    def __init__(self, name: str = "") -> None: ...
    def name(self) -> str: ...

class MOSFETParams:
    vth: float
    kp: float
    lambda_: float
    def __init__(self) -> None: ...

class MOSFET:
    def __init__(self, params: MOSFETParams, name: str = "") -> None: ...
    def name(self) -> str: ...

class IGBTParams:
    vce_sat: float
    beta: float
    def __init__(self) -> None: ...

class IGBT:
    def __init__(self, params: IGBTParams, name: str = "") -> None: ...
    def name(self) -> str: ...

class PWMParams:
    frequency: float
    duty_cycle: float
    amplitude: float
    phase: float
    def __init__(self) -> None: ...

class PWMVoltageSource:
    def __init__(self, params: PWMParams, name: str = "") -> None: ...
    def name(self) -> str: ...

class SineParams:
    frequency: float
    amplitude: float
    phase: float
    offset: float
    def __init__(self) -> None: ...

class SineVoltageSource:
    def __init__(self, params: SineParams, name: str = "") -> None: ...
    def name(self) -> str: ...

class RampParams:
    slope: float
    initial_value: float
    def __init__(self) -> None: ...

class RampGenerator:
    def __init__(self, params: RampParams, name: str = "") -> None: ...
    def name(self) -> str: ...

class PulseParams:
    v_low: float
    v_high: float
    t_rise: float
    t_fall: float
    t_on: float
    period: float
    def __init__(self) -> None: ...

class PulseVoltageSource:
    def __init__(self, params: PulseParams, name: str = "") -> None: ...
    def name(self) -> str: ...

class PIController:
    kp: float
    ki: float
    def __init__(self, kp: float = 1.0, ki: float = 0.0, name: str = "") -> None: ...
    def name(self) -> str: ...

class PIDController:
    kp: float
    ki: float
    kd: float
    def __init__(self, kp: float = 1.0, ki: float = 0.0, kd: float = 0.0, name: str = "") -> None: ...
    def name(self) -> str: ...

class Comparator:
    threshold: float
    def __init__(self, threshold: float = 0.0, name: str = "") -> None: ...
    def name(self) -> str: ...

class SampleHold:
    def __init__(self, name: str = "") -> None: ...
    def name(self) -> str: ...

class RateLimiter:
    max_rate: float
    def __init__(self, max_rate: float = 1.0, name: str = "") -> None: ...
    def name(self) -> str: ...

class MovingAverageFilter:
    window_size: int
    def __init__(self, window_size: int = 10, name: str = "") -> None: ...
    def name(self) -> str: ...

class HysteresisController:
    upper: float
    lower: float
    def __init__(self, upper: float = 1.0, lower: float = -1.0, name: str = "") -> None: ...
    def name(self) -> str: ...

class LookupTable1D:
    def __init__(self, x: List[float], y: List[float], name: str = "") -> None: ...
    def name(self) -> str: ...

# =============================================================================
# Solver Configuration
# =============================================================================

class Tolerances:
    voltage_abstol: float
    voltage_reltol: float
    current_abstol: float
    current_reltol: float
    residual_tol: float
    def __init__(self) -> None: ...
    @staticmethod
    def defaults() -> "Tolerances": ...

class NewtonOptions:
    max_iterations: int
    initial_damping: float
    min_damping: float
    auto_damping: bool
    track_history: bool
    check_per_variable: bool
    num_nodes: int
    num_branches: int
    tolerances: Tolerances
    def __init__(self) -> None: ...

class NewtonResult:
    solution: List[float]
    status: SolverStatus
    iterations: int
    final_residual: float
    final_weighted_error: float
    error_message: str
    def __init__(self) -> None: ...
    def success(self) -> bool: ...

class GminConfig:
    initial_gmin: float
    final_gmin: float
    reduction_factor: float
    max_steps: int
    enable_logging: bool
    def __init__(self) -> None: ...
    def required_steps(self) -> int: ...

class SourceSteppingConfig:
    initial_scale: float
    final_scale: float
    initial_step: float
    min_step: float
    max_step: float
    max_steps: int
    max_failures: int
    enable_logging: bool
    def __init__(self) -> None: ...

class PseudoTransientConfig:
    initial_dt: float
    max_dt: float
    min_dt: float
    dt_increase: float
    dt_decrease: float
    convergence_threshold: float
    max_iterations: int
    enable_logging: bool
    def __init__(self) -> None: ...

class InitializationConfig:
    default_voltage: float
    supply_voltage: float
    diode_forward: float
    mosfet_threshold: float
    use_zero_init: bool
    use_warm_start: bool
    max_random_restarts: int
    random_seed: int
    random_voltage_range: float
    def __init__(self) -> None: ...

class DCConvergenceConfig:
    strategy: DCStrategy
    gmin_config: GminConfig
    source_config: SourceSteppingConfig
    pseudo_config: PseudoTransientConfig
    init_config: InitializationConfig
    enable_random_restart: bool
    max_strategy_attempts: int
    def __init__(self) -> None: ...

class DCAnalysisResult:
    newton_result: NewtonResult
    strategy_used: DCStrategy
    random_restarts: int
    total_newton_iterations: int
    success: bool
    message: str
    def __init__(self) -> None: ...

class StiffnessConfig:
    enable: bool
    rejection_streak_threshold: int
    newton_iter_threshold: int
    newton_streak_threshold: int
    cooldown_steps: int
    dt_backoff: float
    max_bdf_order: int
    monitor_conditioning: bool
    conditioning_error_threshold: float
    switch_integrator: bool
    stiff_integrator: Integrator
    def __init__(self) -> None: ...

class PeriodicSteadyStateOptions:
    period: float
    max_iterations: int
    tolerance: float
    relaxation: float
    store_last_transient: bool
    def __init__(self) -> None: ...

class HarmonicBalanceOptions:
    period: float
    num_samples: int
    max_iterations: int
    tolerance: float
    relaxation: float
    initialize_from_transient: bool
    def __init__(self) -> None: ...

class FrequencyAnalysisPort:
    positive_node: str
    negative_node: str
    def __init__(self) -> None: ...

class FrequencyAnalysisOptions:
    enabled: bool
    mode: FrequencyAnalysisMode
    anchor_mode: FrequencyAnchorMode
    sweep_scale: FrequencySweepScale
    f_start_hz: float
    f_stop_hz: float
    points: int
    injection_current_amplitude: float
    perturbation_port: FrequencyAnalysisPort
    output_port: FrequencyAnalysisPort
    def __init__(self) -> None: ...

class AveragedConverterOptions:
    enabled: bool
    topology: AveragedConverterTopology
    operating_mode: AveragedOperatingMode
    envelope_policy: AveragedEnvelopePolicy
    vin_source: str
    inductor: str
    capacitor: str
    load_resistor: str
    output_node: str
    duty: float
    duty_min: float
    duty_max: float
    switching_frequency_hz: float
    initial_inductor_current: float
    initial_output_voltage: float
    ccm_current_threshold: float
    def __init__(self) -> None: ...

class SwitchingEnergy:
    eon: float
    eoff: float
    err: float
    def __init__(self) -> None: ...

class SwitchingEnergySurface3D:
    current_axis: List[float]
    voltage_axis: List[float]
    temperature_axis: List[float]
    eon_table: List[float]
    eoff_table: List[float]
    err_table: List[float]
    def __init__(self) -> None: ...
    def has_eon(self) -> bool: ...
    def has_eoff(self) -> bool: ...
    def has_err(self) -> bool: ...
    def expected_table_size(self) -> int: ...
    def valid_shape(self) -> bool: ...
    def evaluate_eon(self, current: float, voltage: float, temperature: float) -> float: ...
    def evaluate_eoff(self, current: float, voltage: float, temperature: float) -> float: ...
    def evaluate_err(self, current: float, voltage: float, temperature: float) -> float: ...

class LinearSolverKind(Enum):
    SparseLU = ...
    EnhancedSparseLU = ...
    KLU = ...
    GMRES = ...
    BiCGSTAB = ...
    CG = ...

class PreconditionerKind(Enum):
    None_ = ...
    Jacobi = ...
    ILU0 = ...
    ILUT = ...
    AMG = ...

class IterativeSolverConfig:
    max_iterations: int
    tolerance: float
    restart: int
    preconditioner: PreconditionerKind
    enable_scaling: bool
    scaling_floor: float
    ilut_drop_tolerance: float
    ilut_fill_factor: float
    def __init__(self) -> None: ...
    @staticmethod
    def defaults() -> "IterativeSolverConfig": ...

class LinearSolverStackConfig:
    order: List[LinearSolverKind]
    fallback_order: List[LinearSolverKind]
    direct_config: "LinearSolverConfig"
    iterative_config: IterativeSolverConfig
    allow_fallback: bool
    auto_select: bool
    size_threshold: int
    nnz_threshold: int
    diag_min_threshold: float
    def __init__(self) -> None: ...
    @staticmethod
    def defaults() -> "LinearSolverStackConfig": ...

class LinearSolverTelemetry:
    total_solve_calls: int
    total_analyze_calls: int
    total_factorize_calls: int
    total_iterations: int
    total_fallbacks: int
    last_iterations: int
    last_error: float
    total_analyze_time_seconds: float
    total_factorize_time_seconds: float
    total_solve_time_seconds: float
    last_analyze_time_seconds: float
    last_factorize_time_seconds: float
    last_solve_time_seconds: float
    last_solver: Optional[LinearSolverKind]
    last_preconditioner: Optional[PreconditionerKind]
    def __init__(self) -> None: ...

class SimulationEventType(Enum):
    SwitchOn = ...
    SwitchOff = ...
    ConvergenceWarning = ...
    TimestepChange = ...

class FallbackReasonCode(Enum):
    NewtonFailure = ...
    LTERejection = ...
    EventSplit = ...
    StiffnessBackoff = ...
    TransientGminEscalation = ...
    MaxRetriesExceeded = ...

class ThermalCouplingPolicy(Enum):
    LossOnly = ...
    LossWithTemperatureScaling = ...

class ThermalNetworkKind(Enum):
    SingleRC = ...
    Foster = ...
    Cauer = ...

class SimulationEvent:
    time: float
    type: SimulationEventType
    component: str
    description: str
    value1: float
    value2: float
    def __init__(self) -> None: ...

class FallbackPolicyOptions:
    trace_retries: bool
    enable_transient_gmin: bool
    gmin_retry_threshold: int
    gmin_initial: float
    gmin_max: float
    gmin_growth: float
    def __init__(self) -> None: ...

class BackendTelemetry:
    requested_backend: str
    selected_backend: str
    solver_family: str
    formulation_mode: str
    function_evaluations: int
    jacobian_evaluations: int
    nonlinear_iterations: int
    nonlinear_convergence_failures: int
    error_test_failures: int
    escalation_count: int
    reinitialization_count: int
    backend_recovery_count: int
    state_space_primary_steps: int
    dae_fallback_steps: int
    segment_non_admissible_steps: int
    segment_model_cache_hits: int
    segment_model_cache_misses: int
    linear_factor_cache_hits: int
    linear_factor_cache_misses: int
    equation_assemble_system_calls: int
    equation_assemble_residual_calls: int
    equation_assemble_system_time_seconds: float
    equation_assemble_residual_time_seconds: float
    failure_reason: str
    def __init__(self) -> None: ...

class FallbackTraceEntry:
    step_index: int
    retry_index: int
    time: float
    dt: float
    reason: FallbackReasonCode
    solver_status: SolverStatus
    action: str
    def __init__(self) -> None: ...

class ThermalCouplingOptions:
    enable: bool
    ambient: float
    policy: ThermalCouplingPolicy
    default_rth: float
    default_cth: float
    def __init__(self) -> None: ...

class ThermalDeviceConfig:
    enabled: bool
    network_kind: ThermalNetworkKind
    rth: float
    cth: float
    stage_rth: List[float]
    stage_cth: List[float]
    temp_init: float
    temp_ref: float
    alpha: float
    shared_sink_id: str
    shared_sink_rth: float
    shared_sink_cth: float
    def __init__(self) -> None: ...

class DeviceThermalTelemetry:
    device_name: str
    enabled: bool
    final_temperature: float
    peak_temperature: float
    average_temperature: float
    def __init__(self) -> None: ...

class ThermalSummary:
    enabled: bool
    ambient: float
    max_temperature: float
    device_temperatures: List[DeviceThermalTelemetry]
    def __init__(self) -> None: ...

class ComponentElectrothermalTelemetry:
    component_name: str
    thermal_enabled: bool
    conduction: float
    turn_on: float
    turn_off: float
    reverse_recovery: float
    total_loss: float
    total_energy: float
    average_power: float
    peak_power: float
    final_temperature: float
    peak_temperature: float
    average_temperature: float
    def __init__(self) -> None: ...

class SystemLossSummary:
    device_losses: Dict[str, object]
    total_loss: float
    total_conduction: float
    total_switching: float
    input_power: float
    output_power: float
    efficiency: float
    def __init__(self) -> None: ...
    def compute_totals(self) -> None: ...

class SimulationOptions:
    tstart: float
    tstop: float
    dt: float
    dt_min: float
    dt_max: float
    newton_options: NewtonOptions
    dc_config: DCConvergenceConfig
    linear_solver: LinearSolverStackConfig
    adaptive_timestep: bool
    step_mode: StepMode
    control_mode: ControlUpdateMode
    control_sample_time: float
    formulation_mode: FormulationMode
    direct_formulation_fallback: bool
    timestep_config: "AdvancedTimestepConfig"
    lte_config: "RichardsonLTEConfig"
    integrator: Integrator
    enable_bdf_order_control: bool
    bdf_config: "BDFOrderConfig"
    stiffness_config: StiffnessConfig
    enable_periodic_shooting: bool
    periodic_options: PeriodicSteadyStateOptions
    enable_harmonic_balance: bool
    harmonic_balance: HarmonicBalanceOptions
    frequency_analysis: FrequencyAnalysisOptions
    averaged_converter: AveragedConverterOptions
    enable_events: bool
    enable_losses: bool
    switching_energy: Dict[str, SwitchingEnergy]
    switching_energy_surfaces: Dict[str, SwitchingEnergySurface3D]
    thermal: ThermalCouplingOptions
    thermal_devices: Dict[str, ThermalDeviceConfig]
    gmin_fallback: GminConfig
    max_step_retries: int
    fallback_policy: FallbackPolicyOptions
    def __init__(self) -> None: ...

class SimulationResult:
    time: List[float]
    states: List[List[float]]
    events: List[SimulationEvent]
    mixed_domain_phase_order: List[str]
    virtual_channels: Dict[str, List[float]]
    virtual_channel_metadata: Dict[str, VirtualChannelMetadata]
    success: bool
    final_status: SolverStatus
    message: str
    total_steps: int
    newton_iterations_total: int
    timestep_rejections: int
    total_time_seconds: float
    linear_solver_telemetry: LinearSolverTelemetry
    fallback_trace: List[FallbackTraceEntry]
    backend_telemetry: BackendTelemetry
    loss_summary: SystemLossSummary
    thermal_summary: ThermalSummary
    component_electrothermal: List[ComponentElectrothermalTelemetry]
    data: List[List[float]]
    def __init__(self) -> None: ...

class PeriodicSteadyStateResult:
    success: bool
    iterations: int
    residual_norm: float
    steady_state: List[float]
    last_cycle: SimulationResult
    message: str
    def __init__(self) -> None: ...

class HarmonicBalanceResult:
    success: bool
    iterations: int
    residual_norm: float
    solution: List[float]
    sample_times: List[float]
    message: str
    def __init__(self) -> None: ...

class FrequencyAnalysisResult:
    success: bool
    diagnostic: object
    message: str
    mode: FrequencyAnalysisMode
    anchor_mode_selected: FrequencyAnchorMode
    failed_point_index: int
    failed_frequency_hz: float
    frequency_hz: List[float]
    response_real: List[float]
    response_imag: List[float]
    magnitude: List[float]
    magnitude_db: List[float]
    phase_deg: List[float]
    channel_metadata: Dict[str, VirtualChannelMetadata]
    gain_crossover_hz: float
    phase_crossover_hz: float
    phase_margin_deg: float
    gain_margin_db: float
    gain_crossover_reason: FrequencyMetricUndefinedReason
    phase_crossover_reason: FrequencyMetricUndefinedReason
    phase_margin_reason: FrequencyMetricUndefinedReason
    gain_margin_reason: FrequencyMetricUndefinedReason
    def __init__(self) -> None: ...

class Simulator:
    options: SimulationOptions
    def __init__(self, circuit: Circuit, options: SimulationOptions = ...) -> None: ...
    def dc_operating_point(self) -> DCAnalysisResult: ...
    def run_transient(self, x0: Optional[List[float]] = ...) -> SimulationResult: ...
    def run_periodic_shooting(
        self,
        x0_or_options: Optional[object] = ...,
        options: Optional[PeriodicSteadyStateOptions] = ...,
    ) -> PeriodicSteadyStateResult: ...
    def run_harmonic_balance(
        self,
        x0_or_options: Optional[object] = ...,
        options: Optional[HarmonicBalanceOptions] = ...,
    ) -> HarmonicBalanceResult: ...
    def run_frequency_analysis(self, options: Optional[FrequencyAnalysisOptions] = ...) -> FrequencyAnalysisResult: ...
    def set_switching_energy(self, device_name: str, energy: SwitchingEnergy) -> None: ...
    def set_switching_energy_surface(self, device_name: str, surface: SwitchingEnergySurface3D) -> None: ...

class YamlParserOptions:
    strict: bool
    validate_nodes: bool
    def __init__(self) -> None: ...

class YamlParser:
    errors: List[str]
    warnings: List[str]
    def __init__(self, options: YamlParserOptions = ...) -> None: ...
    def load(self, path: str) -> Tuple[Circuit, SimulationOptions]: ...
    def load_string(self, content: str) -> Tuple[Circuit, SimulationOptions]: ...

# =============================================================================
# Validation Framework
# =============================================================================

class RCAnalytical:
    def __init__(self, R: float, C: float, V_initial: float, V_final: float) -> None: ...
    def tau(self) -> float: ...
    def voltage(self, t: float) -> float: ...
    def current(self, t: float) -> float: ...
    def waveform(self, t_start: float, t_end: float, dt: float) -> List[float]: ...

class RLAnalytical:
    def __init__(self, R: float, L: float, V_source: float, I_initial: float) -> None: ...
    def tau(self) -> float: ...
    def I_final(self) -> float: ...
    def current(self, t: float) -> float: ...
    def voltage_R(self, t: float) -> float: ...
    def voltage_L(self, t: float) -> float: ...
    def waveform(self, t_start: float, t_end: float, dt: float) -> List[float]: ...

class RLCAnalytical:
    def __init__(self, R: float, L: float, C: float, V_source: float, V_initial: float, I_initial: float) -> None: ...
    def omega_0(self) -> float: ...
    def zeta(self) -> float: ...
    def alpha(self) -> float: ...
    def damping_type(self) -> RLCDamping: ...
    def voltage(self, t: float) -> float: ...
    def current(self, t: float) -> float: ...
    def waveform(self, t_start: float, t_end: float, dt: float) -> List[float]: ...

class ValidationResult_v2:
    test_name: str
    passed: bool
    max_error: float
    rms_error: float
    max_relative_error: float
    mean_error: float
    num_points: int
    error_threshold: float
    def __init__(self) -> None: ...
    def to_string(self) -> str: ...

def compare_waveforms(name: str, simulated: List[tuple], analytical: List[tuple], threshold: float = 0.001) -> ValidationResult_v2: ...
def export_validation_csv(results: List[ValidationResult_v2]) -> str: ...
def export_validation_json(results: List[ValidationResult_v2]) -> str: ...

# =============================================================================
# Benchmark Framework
# =============================================================================

class BenchmarkTiming:
    name: str
    iterations: int
    def __init__(self) -> None: ...
    def average_ms(self) -> float: ...
    def min_ms(self) -> float: ...
    def max_ms(self) -> float: ...
    def total_ms(self) -> float: ...

class BenchmarkResult:
    circuit_name: str
    num_nodes: int
    num_devices: int
    num_timesteps: int
    timing: BenchmarkTiming
    simulation_time: float
    def __init__(self) -> None: ...
    def timesteps_per_second(self) -> float: ...
    def to_string(self) -> str: ...

def export_benchmark_csv(results: List[BenchmarkResult]) -> str: ...
def export_benchmark_json(results: List[BenchmarkResult]) -> str: ...

# =============================================================================
# Integration Methods
# =============================================================================

class BDFOrderConfig:
    min_order: int
    max_order: int
    initial_order: int
    order_increase_threshold: float
    order_decrease_threshold: float
    steps_before_increase: int
    enable_auto_order: bool
    def __init__(self) -> None: ...

class TimestepConfig:
    dt_min: float
    dt_max: float
    dt_initial: float
    safety_factor: float
    error_tolerance: float
    growth_factor: float
    shrink_factor: float
    max_rejections: int
    k_p: float
    k_i: float
    def __init__(self) -> None: ...
    @staticmethod
    def defaults() -> "TimestepConfig": ...
    @staticmethod
    def conservative() -> "TimestepConfig": ...
    @staticmethod
    def aggressive() -> "TimestepConfig": ...

class AdvancedTimestepConfig(TimestepConfig):
    target_newton_iterations: int
    min_newton_iterations: int
    max_newton_iterations: int
    newton_feedback_gain: float
    max_growth_rate: float
    max_shrink_rate: float
    enable_smoothing: bool
    lte_weight: float
    newton_weight: float
    def __init__(self) -> None: ...
    @staticmethod
    def defaults() -> "AdvancedTimestepConfig": ...
    @staticmethod
    def for_switching() -> "AdvancedTimestepConfig": ...
    @staticmethod
    def for_power_electronics() -> "AdvancedTimestepConfig": ...

class RichardsonLTEConfig:
    method: TimestepMethod
    extrapolation_order: int
    voltage_tolerance: float
    current_tolerance: float
    use_weighted_norm: bool
    history_depth: int
    def __init__(self) -> None: ...
    @staticmethod
    def defaults() -> "RichardsonLTEConfig": ...
    @staticmethod
    def step_doubling() -> "RichardsonLTEConfig": ...

# =============================================================================
# High-Performance Features
# =============================================================================

class LinearSolverConfig:
    pivot_tolerance: float
    reuse_symbolic: bool
    detect_pattern_change: bool
    deterministic_pivoting: bool
    def __init__(self) -> None: ...

def detect_simd_level() -> SIMDLevel: ...
def simd_vector_width() -> int: ...
def backend_capabilities() -> Dict[str, bool]: ...
def solver_status_to_string(status: SolverStatus) -> str: ...
def solve_dc(circuit: Circuit, options: Optional[NewtonOptions] = None) -> NewtonResult: ...
def dc_operating_point(circuit: Circuit, options: Optional[DCConvergenceConfig] = None) -> DCAnalysisResult: ...
def run_transient(
    circuit: Circuit,
    t_start: float,
    t_stop: float,
    dt: float,
    x0: Optional[List[float]] = None,
    newton_options: Optional[NewtonOptions] = None,
    linear_solver: Optional[LinearSolverStackConfig] = None,
) -> Tuple[List[float], List[List[float]], bool, str]: ...
def run_transient_streaming(
    circuit: Circuit,
    t_start: float,
    t_stop: float,
    dt: float,
    x0: Optional[List[float]] = None,
    newton_options: Optional[NewtonOptions] = None,
    linear_solver: Optional[LinearSolverStackConfig] = None,
) -> object: ...
def run_frequency_analysis(
    circuit: Circuit,
    options: FrequencyAnalysisOptions,
) -> FrequencyAnalysisResult: ...

# =============================================================================
# Thermal & Loss Models
# =============================================================================

class FosterStage:
    r: float
    c: float
    def __init__(self) -> None: ...

class FosterNetwork:
    stages: List[FosterStage]
    def __init__(self) -> None: ...

class CauerStage:
    r: float
    c: float
    def __init__(self) -> None: ...

class CauerNetwork:
    stages: List[CauerStage]
    def __init__(self) -> None: ...

class ThermalSimulator:
    def __init__(self) -> None: ...

class ThermalLimitMonitor:
    def __init__(self) -> None: ...

class ThermalResult:
    def __init__(self) -> None: ...

def create_mosfet_thermal_model() -> FosterNetwork: ...
def create_from_datasheet_4param(rth_jc: float, rth_cs: float, rth_sa: float, cth_j: float) -> FosterNetwork: ...
def create_simple_thermal_model(rth: float, cth: float) -> FosterNetwork: ...

class MOSFETLossParams:
    def __init__(self) -> None: ...

class IGBTLossParams:
    def __init__(self) -> None: ...

class DiodeLossParams:
    def __init__(self) -> None: ...

class ConductionLoss:
    def __init__(self) -> None: ...

class SwitchingLoss:
    def __init__(self) -> None: ...

class LossBreakdown:
    def __init__(self) -> None: ...

class LossAccumulator:
    def __init__(self) -> None: ...

class EfficiencyCalculator:
    def __init__(self) -> None: ...

class LossResult:
    def __init__(self) -> None: ...

# =============================================================================
# Convergence Monitoring
# =============================================================================

class IterationRecord:
    def __init__(self) -> None: ...

class ConvergenceHistory:
    def __init__(self) -> None: ...

class VariableConvergence:
    def __init__(self) -> None: ...

class PerVariableConvergence:
    def __init__(self) -> None: ...



"""Example 01 — Python gain block via PythonCBlock.

Demonstrates using PythonCBlock to double an input signal without any
C compiler, running inside a SignalEvaluator pipeline.

Run::

    PYTHONPATH=python python examples/cblock/01_python_gain.py
"""

from pulsim.cblock import PythonCBlock

# --- Define a gain-of-2 block in Python ------------------------------------

def gain2(ctx, t, dt, inputs):
    """Double the first input signal."""
    return [2.0 * inputs[0]]


blk = PythonCBlock(gain2, n_inputs=1, n_outputs=1, name="gain2")

# Simulate 5 steps
for i in range(5):
    t = i * 1e-4
    dt = 1e-4 if i > 0 else 0.0
    out = blk.step(t, dt, [float(i)])
    print(f"t={t:.4f}s  in={float(i):.1f}  out={out[0]:.1f}")

# Expected output:
# t=0.0000s  in=0.0  out=0.0
# t=0.0001s  in=1.0  out=2.0
# t=0.0002s  in=2.0  out=4.0
# t=0.0003s  in=3.0  out=6.0
# t=0.0004s  in=4.0  out=8.0



#!/usr/bin/env python3
"""Update project version across release-managed files.

Usage:
    python scripts/update_version.py 0.3.4
    python scripts/update_version.py v0.3.4 --dry-run
"""

from __future__ import annotations

import argparse
import re
import sys
from dataclasses import dataclass
from pathlib import Path


ROOT = Path(__file__).resolve().parents[1]
SEMVER_RE = re.compile(r"^\d+\.\d+\.\d+$")


@dataclass(frozen=True)
class Rule:
    pattern: re.Pattern[str]
    replacement: str
    expected_count: int = 1
    description: str = ""


RULES: dict[str, list[Rule]] = {
    "pyproject.toml": [
        Rule(
            pattern=re.compile(r'(?m)^version\s*=\s*"[^"]+"$'),
            replacement='version = "{version}"',
            description="Python package version",
        ),
    ],
    "CMakeLists.txt": [
        Rule(
            pattern=re.compile(
                r"(?m)^project\(pulsim VERSION \d+\.\d+\.\d+ LANGUAGES CXX\)$"
            ),
            replacement="project(pulsim VERSION {version} LANGUAGES CXX)",
            description="CMake project version",
        ),
    ],
    "python/pulsim/__init__.py": [
        Rule(
            pattern=re.compile(r'(?m)^__version__\s*=\s*"[^"]+"$'),
            replacement='__version__ = "{version}"',
            description="Python runtime __version__",
        ),
    ],
    "docs/python/conf.py": [
        Rule(
            pattern=re.compile(r"(?m)^version\s*=\s*'[^']+'$"),
            replacement="version = '{version}'",
            description="Sphinx short version",
        ),
        Rule(
            pattern=re.compile(r"(?m)^release\s*=\s*'[^']+'$"),
            replacement="release = '{version}'",
            description="Sphinx release version",
        ),
    ],
    "docs/Doxyfile": [
        Rule(
            pattern=re.compile(r'(?m)^PROJECT_NUMBER\s*=\s*"[^"]*"$'),
            replacement='PROJECT_NUMBER         = "{version}"',
            description="Doxygen project number",
        ),
    ],
}


def normalize_version(raw: str) -> str:
    version = raw[1:] if raw.startswith("v") else raw
    if not SEMVER_RE.fullmatch(version):
        raise ValueError(
            f"Invalid version '{raw}'. Expected format X.Y.Z (optionally prefixed with 'v')."
        )
    return version


def apply_rules(path: Path, rules: list[Rule], version: str) -> tuple[bool, str]:
    content = path.read_text(encoding="utf-8")
    original = content
    for rule in rules:
        replacement = rule.replacement.format(version=version)
        content, count = rule.pattern.subn(replacement, content)
        if count != rule.expected_count:
            desc = rule.description or rule.pattern.pattern
            raise RuntimeError(
                f"{path}: expected {rule.expected_count} replacement(s) for '{desc}', got {count}."
            )
    return content != original, content


def main() -> int:
    parser = argparse.ArgumentParser(
        description="Update release-managed version strings across project files."
    )
    parser.add_argument("version", help="Target version (X.Y.Z or vX.Y.Z)")
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Show what would change without writing files.",
    )
    args = parser.parse_args()

    try:
        version = normalize_version(args.version)
    except ValueError as exc:
        print(f"ERROR: {exc}", file=sys.stderr)
        return 2

    changed_files: list[str] = []
    touched_rules = 0

    for rel_path, rules in RULES.items():
        path = ROOT / rel_path
        if not path.exists():
            print(f"ERROR: Missing file: {path}", file=sys.stderr)
            return 1

        try:
            changed, updated_content = apply_rules(path, rules, version)
        except RuntimeError as exc:
            print(f"ERROR: {exc}", file=sys.stderr)
            return 1

        touched_rules += len(rules)
        if changed:
            changed_files.append(rel_path)
            if not args.dry_run:
                path.write_text(updated_content, encoding="utf-8")

    mode = "DRY-RUN" if args.dry_run else "UPDATED"
    print(f"{mode}: set version to {version}")
    print(f"Checked {len(RULES)} files / {touched_rules} rules.")
    if changed_files:
        for rel_path in changed_files:
            print(f" - {rel_path}")
    else:
        print("No file content changes were necessary.")

    return 0


if __name__ == "__main__":
    raise SystemExit(main())
